package be.iminds.ilabt.jfed.testing.base;

import be.iminds.ilabt.jfed.call_log_output.LogOutput;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.TargetAuthority;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * ApiTest:
 *
 *   for a test class you need:
 *     - user info
 *     - SA info if applicable
 *     - tested API info
 *     - connection (auto)
 *     - logger (auto)
 *
 *  per test meta data
 *     - name (default: auto from method name)
 *     - dependencies and preferred grouping (name or methodName)
 *        - hard dependency (skip if other failed or skipped)
 *        - soft dependency (always run after other (no matter if it succeeds, skips or fails))
 *        - group (a name that can be the same for a number of tests. These will be kept together, unless the other dependencies are used)
 *
 *  for each test log:
 *     - commands and replies
 *     - details, notes, warnings, errors and fatal errors (+ skips: the test can start to run and then decide it needs to be skipped)
 *     - info on the test
 *         - description of what it does
 *         - description of what is expected
 */
public abstract class ApiTest {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(ApiTest.class);

    protected final Logger logger;
    protected final TargetAuthority testedAuthority;
    protected final GeniUser user;

    protected final JFedConnectionProvider connectionProvider;
    protected final AuthorityListModel authorityListModel;
    protected final AuthorityFinder authorityFinder;
    protected final JFedPreferences jFedPreferences;


    public Logger getLogger() {
        return logger;
    }
    public TargetAuthority getTestedAuthority() {
        return testedAuthority;
    }
    public GeniUser getUser() {
        return user;
    }
    public JFedConnectionProvider getConnectionProvider() {
        return connectionProvider;
    }
    public AuthorityListModel getAuthorityListModel() {
        return authorityListModel;
    }
    public GeniUserProvider getGeniUserProvider() {
        return new GeniUserProvider() {
            @Override
            public GeniUser getLoggedInGeniUser() {
                return user;
            }

            @Override
            public boolean isUserLoggedIn() {
                return user != null;
            }
        };
    }
    public AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory getAutomaticUserAndSliceApiWrapperFactory() {
        return new AutomaticUserAndSliceApiWrapper.BasicAutomaticUserAndSliceApiWrapperFactory(logger, user, connectionProvider, jFedPreferences);
    }
    public AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory getAutomaticAggregateManagerWrapperFactory() {
        return new AutomaticAggregateManagerWrapper.BasicAutomaticAggregateManagerWrapperFactory(logger, user, connectionProvider, jFedPreferences);
    }


    public ApiTest(Logger logger, TargetAuthority testedAuthority, GeniUser user, JFedConnectionProvider connectionProvider,
                   AuthorityListModel authorityListModel, JFedPreferences jFedPreferences) {
        this.logger = logger;
        this.testedAuthority = testedAuthority;
        this.user = user;
        this.connectionProvider = connectionProvider;
        this.authorityListModel = authorityListModel;
        this.jFedPreferences = jFedPreferences;
        this.authorityFinder = new AuthorityFinder(authorityListModel, jFedPreferences); //safe fallback as AuthorityFinder does not keep state
    }
    public ApiTest(Logger logger, TargetAuthority testedAuthority, GeniUser user, JFedConnectionProvider connectionProvider,
                   AuthorityListModel authorityListModel, AuthorityFinder authorityFinder, JFedPreferences jFedPreferences) {
        this.logger = logger;
        this.testedAuthority = testedAuthority;
        this.user = user;
        this.connectionProvider = connectionProvider;
        this.authorityListModel = authorityListModel;
        this.authorityFinder = authorityFinder;
        this.jFedPreferences = jFedPreferences;
    }

    private Properties testConfig;
    public Properties getTestConfig() {
        return testConfig;
    }
    void setTestConfig(Properties testConfig) {
        this.testConfig = testConfig;
    }

    /** Overwrite to specify required config keys */
    protected List<String> getReqConfKeys() {
        return new ArrayList<String>();
    }
    /** Overwrite to specify optional config keys */
    protected List<String> getOptConfKeys() {
        return new ArrayList<String>();
    }


    protected abstract String getTestDescription();
    public abstract void setUp() throws Exception;

    protected ApiTestResult.ApiTestMethodResult currentTestResult;
    void setCurrentTestResult(ApiTestResult.ApiTestMethodResult currentTestResult) {
        this.currentTestResult = currentTestResult;
    }
    ApiTestResult.ApiTestMethodResult getCurrentTestResult() {
        return currentTestResult;
    }
    public LogOutput.TestResultState getCurrentResultState() {
        return currentTestResult.getState();
    }

    public JFedPreferences getJFedPreferences() {
        return jFedPreferences;
    }

    //TODO Replace these 3 with a description and options defined using class annotations?
    public static String getDescription(Class<? extends ApiTest> testClass) {
        //bit of a hack, but it works with all null's ...
        ApiTestFactory apiTestFactory = new ApiTestFactory(null/*injector*/, null/*user*/, null/* connectionProvider*/, null/* authorityListModel*/, null/* jFedPreferences*/, null/*automaticAggregateManagerWrapperFactory*/);
        try {
            ApiTest test = apiTestFactory.createNullApiTest(testClass);
            return test.getTestDescription();
        } catch (ApiTestFactory.InvalidApiTestException e) {
            LOG.error("Error fetching test description", e);
            return "Error fetching test description";
        }
    }
    public static List<String> getRequiredConfigKeys(Class<? extends ApiTest> testClass) {
        ApiTestFactory apiTestFactory = new ApiTestFactory(null/*injector*/, null/*user*/, null/* connectionProvider*/, null/* authorityListModel*/, null/* jFedPreferences*/, null/*automaticAggregateManagerWrapperFactory*/);
        try {
            ApiTest test = apiTestFactory.createNullApiTest(testClass);
            return test.getReqConfKeys();
        } catch (ApiTestFactory.InvalidApiTestException e) {
            LOG.error("Error fetching test description", e);
            return new ArrayList<String>();
        }
    }
    public static List<String> getOptionalConfigKeys(Class<? extends ApiTest> testClass) {
        ApiTestFactory apiTestFactory = new ApiTestFactory(null/*injector*/, null/*user*/, null/* connectionProvider*/, null/* authorityListModel*/, null/* jFedPreferences*/, null/*automaticAggregateManagerWrapperFactory*/);
        try {
            ApiTest test = apiTestFactory.createNullApiTest(testClass);
            return test.getOptConfKeys();
        } catch (ApiTestFactory.InvalidApiTestException e) {
            LOG.error("Error fetching test description", e);
            return new ArrayList<String>();
        }
    }

    @java.lang.annotation.Retention(java.lang.annotation.RetentionPolicy.RUNTIME)
    @java.lang.annotation.Target({java.lang.annotation.ElementType.METHOD})
    public @interface Test {
        java.lang.String description() default "";
        java.lang.String[] groups() default {};

        java.lang.String[] hardDepends() default {};
        java.lang.String[] softDepends() default {};

        String dataProvider() default ""; //TODO implement

        //this concept (isReturnValue annotation) is under revision
//        /**
//         * If true, only the success of this test will be used as return value for the entire test run
//         * If there are no tests with this option, the success of all tests is used.
//         * If there are multiple tests in a class with this option, the result is unspecified!
//         * */
//        boolean ignoreForReturnValue() default false;
    }
    public @interface DataProvider { //TODO implement
        String name() default "";
    }


    public void note(String val) {
        currentTestResult.addLogLine(LogOutput.LogLineType.NOTE, val);
    }
    public void note(String val, Throwable t) {
        currentTestResult.addLogLine(LogOutput.LogLineType.NOTE, val, t);
    }

    public void skip(String reason) {
        currentTestResult.state = ApiTestResult.TestResultState.SKIPPED;
//        currentTestResult.addLogLine(ApiTestResult.MethodLogLineType.NOTE, "Test skipped: "+reason);
        throw new AutomatedTestRunner.SkipTestException(reason);
    }

    public void skip(String reason, Throwable t) {
        currentTestResult.state = ApiTestResult.TestResultState.SKIPPED;
        throw new AutomatedTestRunner.SkipTestException(reason, t);
    }
    private boolean errorsFatal = true;
    private boolean errorsAreWarnings = false;
    /** errors will no longer be fatal, until setErrorsFatal() is called, or the current annotated test is done (so next test methods wil have fatal errors again). */
    public void setErrorsNotFatal() {
        errorsFatal = false;
        errorsAreWarnings = false;
    }
    public void setErrorsFatal() {
        errorsFatal = true;
        errorsAreWarnings = false;
    }

    /**
     * Errors are no longer errors, and are not fatal. They are warnings instead.
     * This works until setErrorsAreWarnings(false), setErrorsNotFatal() or setErrorsFatal() is called, or until the current annotated test is done (so next test methods wil have fatal errors again).
     * */
    public void setErrorsAreWarnings(boolean warn) {
        errorsAreWarnings = warn;
    }


    /**
     * Makes the test fail, but continues executing the test code.
     * @param val text describing the error
     */
    public void errorNonFatal(String val) {
        currentTestResult.state = ApiTestResult.TestResultState.FAILED;
        currentTestResult.addLogLine(LogOutput.LogLineType.ERROR, val);
    }
    public void errorNonFatal(String val, Throwable t) {
        currentTestResult.state = ApiTestResult.TestResultState.FAILED;
        currentTestResult.addLogLine(LogOutput.LogLineType.ERROR, val ,t);
    }

    public void errorFatal(String val) {
        fatalError(val);
    }
    public void errorFatal(String val, Throwable t) {
        fatalError(val, t);
    }
    public void fatalError(String val) {
        fatalError(val, null);
    }
    public void fatalError(String val, Throwable t) {
        if (errorsAreWarnings) {
            warn(val, t);
        } else {
            if (errorsFatal) {
//        currentTestResult.addLogLine(ApiTestResult.MethodLogLineType.FATAL_ERROR, val);
                currentTestResult.state = ApiTestResult.TestResultState.FAILED;
                throw new AssertionError(val, t);
            } else {
                errorNonFatal(val, t);
            }
        }
    }

    public void warn(String val) {
        currentTestResult.addLogLine(LogOutput.LogLineType.WARN, val);
        if (currentTestResult.state == ApiTestResult.TestResultState.SUCCESS)
            currentTestResult.state = ApiTestResult.TestResultState.WARN;
    }
    public void warn(String val, Throwable e) {
        currentTestResult.addLogLine(LogOutput.LogLineType.WARN, val, e);
        if (currentTestResult.state == ApiTestResult.TestResultState.SUCCESS)
            currentTestResult.state = ApiTestResult.TestResultState.WARN;
    }

    public void warnIfNot(boolean v) {
        warnIfNot(v, "warnIfNot failed");
    }
    public void warnIfNot(boolean v, String failText) {
        if (!v)
            warn(failText);
    }

    public void errorNonFatalIfNot(boolean v) {
        errorNonFatalIfNot(v, "warnIfNot failed");
    }
    public void errorNonFatalIfNot(boolean v, String failText) {
        if (!v)
            errorNonFatal(failText);
    }

    public void assertTrue(boolean v) {
        assertTrue(v, "assertTrue failed");
    }
    public void assertTrue(boolean v, String failText) {
        if (!v)
            fatalError(failText);
    }


    public void assertFalse(boolean v) {
        assertFalse(v, "assertFalse failed");
    }
    public void assertFalse(boolean v, String failText) {
        if (v)
            fatalError(failText);
    }


    public void assertNotNull(Object o) {
        assertNotNull(o, "object is null");
    }
    public void assertNotNull(Object o, String failText) {
        if (o == null)
            fatalError(failText);
    }


    public void assertNonEmptyString(String s) {
        assertNonEmptyString(s, "string should not be empty");
    }
    public void assertNonEmptyString(String s, String failText) {
        if (s == null)
            fatalError(failText);
        if (s.trim().isEmpty())
            fatalError(failText);
    }
    public void assertNotEmpty(Collection c) {
        assertNotEmpty(c, "list is empty");
    }
    public void assertEmpty(Collection c) {
        assertEmpty(c, "list is not empty");
    }
    public void assertNotEmpty(Collection c, String failText) {
        if (c.isEmpty())
            fatalError(failText);
    }
    public void assertEmpty(Collection c, String failText) {
        if (!c.isEmpty())
            fatalError(failText);
    }


    public void assertNull(Object o) {
        assertNotNull(o, "object is null");
    }
    public void assertNull(Object o, String failText) {
        if (o != null)
            fatalError(failText);
    }

    public void assertEquals(Object a, Object b, String failText) {
        if (!a.equals(b))
            fatalError(failText+" -> "+(a instanceof String ? "object are not equal: a=\""+a.toString()+"\" b=\""+
                    b.toString()+"\"": "object are not equal: a="+a.toString() +"b="+b.toString()));
    }
    public void assertEquals(Object a, Object b) {
        assertEquals(a, b, a instanceof String ? "object are not equal: a=\""+a.toString()+"\" b=\""
                +b.toString()+"\"": "object are not equal: a="+a.toString() +"b="+b.toString());
    }
    /** assert that 2 dates are equal, to second precision. sub seconds precision may differ*/
    public void assertEqualDatesNoMs(Date a, Date b) {
        long as = a.getTime() / 1000l;
        long bs = b.getTime() / 1000l;
        LOG.debug("assertEqualDatesNoMs compare as="+as+" bs="+bs+"   a="+a+"  b="+b);
        if (as != bs)
            fatalError("dates are not equal: "+a.toString()+" differs from " +b.toString()+"");
    }


    public void assertNotEquals(Object a, Object b, String failText) {
        if (a.equals(b))
            fatalError(failText+" -> object are equal: a=b="+a.toString());
    }
    public void assertNotEquals(Object a, Object b) {
        assertNotEquals(a, b, "object are equal: a=b="+a.toString());
    }

    public void assertInstanceOf(Object o, Class<?> c) {
        assertInstanceOf(o, c, "object is not an instance of class \""+c.getName()+"\". (It is instance of class "+c.getName().toString()+")");
    }
    public void assertInstanceOf(Object o, Class<?> c, String failText) {
        if (!c.isInstance(o))
            fatalError(failText+" -> object is not an instance of class \""+c.getName()+"\". (It is instance of class "+c.getName().toString()+")");
    }



    public Vector assertHashTableContainsVector(Hashtable t, String key) {
        Object o = t.get(key);
        assertNotNull(o, "Hashtable does not contain key \""+key+"\"");
        if (o == null) return null;
        assertEquals(o.getClass(), Vector.class, "value for " + key + " is not a Vector but a " + o.getClass().getName());
        if (!o.getClass().equals(Vector.class)) return null;
        return (Vector) o;
    }
    public Hashtable assertHashTableContainsHashtable(Hashtable t, String key) {
        Object o = t.get(key);
        assertNotNull(o, "Hashtable does not contain key \""+key+"\"");
        if (o == null) return null;
        assertEquals(o.getClass(), Hashtable.class, "value for " + key + " is not a Hashtable but a " + o.getClass().getName());
        if (!o.getClass().equals(Hashtable.class)) return null;
        return (Hashtable) o;
    }
    public String assertHashTableContainsString(Hashtable t, String key) {
        Object o = t.get(key);
        assertNotNull(o, "Hashtable does not contain key \""+key+"\"");
        if (o == null) return null;
        assertEquals(o.getClass(), String.class, "value for " + key + " is not a String but a " + o.getClass().getName());
        if (!o.getClass().equals(String.class)) return null;
        return (String) o;
    }
    public String assertHashTableContainsString(Hashtable t, String key, String expectedValue) {
        Object o = t.get(key);
        assertNotNull(o, "Hashtable does not contain key \""+key+"\"");
        if (o == null) return null;
        assertEquals(o.getClass(), String.class, "value for " + key + " is not a String but a " + o.getClass().getName());
        assertEquals(o, expectedValue, "value for "+key+" is not \""+expectedValue+"\" but \""+o+"\"");
        if (!o.getClass().equals(String.class)) return null;
        return (String) o;
    }
    public Object assertHashTableContainsObject(Hashtable t, String key) {
        Object o = t.get(key);
        assertNotNull(o, "Hashtable does not contain key \""+key+"\"");
        return o;
    }
    public String assertHashTableContainsNonemptyString(Hashtable t, String key) {
        Object o = t.get(key);
        assertNotNull(o, "Hashtable does not contain key \""+key+"\"");
        if (o == null) return null;
        assertEquals(o.getClass(), String.class, "value for " + key + " is not a String but a " + o.getClass().getName());
        if (!o.getClass().equals(String.class)) return null;
        String s = (String) o;
        assertTrue(s.length() > 0, "value for " + key + " is an empty string");
        return s;
    }



    /** @param type is allowed to be null if no type to test */
    public void assertValidUrn(String urn, String type) {
        assertValidUrn(urn, type, null);

    }
    public void assertValidUrn(String urn, String type, String message) {
        String msgHelp = message == null ? "" : message+": ";
        GeniUrn geniUrn = GeniUrn.parse(urn);
        assertTrue(geniUrn != null, msgHelp+"Urn is not valid: \"" + urn + "\"");
        if (geniUrn == null) return;
        if (type != null) assertEquals(geniUrn.getEncodedResourceType(), type, msgHelp+"Urn type is not correct in urn=\"" + urn + "\"");
        assertFalse(geniUrn.getEncodedTopLevelAuthority().isEmpty(), msgHelp+"Urn authority is empty in urn=\"" + urn + "\"");
        assertFalse(geniUrn.getEncodedResourceName().isEmpty(), msgHelp+"Urn resource name is empty in urn=\"" + urn + "\"");
    }

}

