package be.iminds.ilabt.jfed.util;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnectionPool;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.stitching.VlanRange;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.UserLoginModelManager;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import javafx.concurrent.Task;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Singleton;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketTimeoutException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * ProxyPreferencesManager
 * <p/>
 * This class sets the preferences from PreferencesUtil, combined with data about failed connections
 * to the ProxySettings in an SfaConnectionPool
 */
@Singleton
public class ProxyPreferencesManager {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ProxyPreferencesManager.class);

    public static final String ALWAYS_USE_PROXY = "ALWAYS";
    public static final String NEVER_USE_PROXY = "NEVER";
    public static final String AUTO_USE_PROXY = "AUTO";

    private final GeniUserProvider userLoginModelManager;
    private final SfaConnectionPool connectionPool;
    private final JFedGuiPreferences jFedPreferences;

    private boolean needProxyForJFed;
    private boolean needProxyForSsh;

    private GeniUser prevUser;
    private JFedConnection.ProxyInfo prevProxy;
    private boolean prevNeedProxyForJFed = false;
//    private boolean prevNeedProxyForSsh = false;


    @Inject
    public ProxyPreferencesManager(UserLoginModelManager userLoginModelManager, SfaConnectionPool connectionPool, JFedGuiPreferences jFedPreferences) {
        this.userLoginModelManager = userLoginModelManager;
        this.connectionPool = connectionPool;
        this.jFedPreferences = jFedPreferences;

//        userLoginModelManager.addUserLoginListener(new UserLoginModelManager.UserLoginListener() {
//            @Override
//            public void onUserLogin(UserLoginModelManager userLoginModel, boolean valid) {
//                updateProxySettings();
//            }
//        });
        updateProxySettings();
    }

    public static class ProxyPort {
        private final SfaAuthority.ProxyInfo proxyInfo;
        private final int port;

        private ProxyPort(SfaAuthority.ProxyInfo proxyInfo, int port) {
            assert proxyInfo != null;
            this.port = port;
            this.proxyInfo = proxyInfo;
        }

        public SfaAuthority.ProxyInfo getProxyInfo() {
            return proxyInfo;
        }

        public String getHostname() {
            return proxyInfo.getHostname();
        }

        public int getPort() {
            return port;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ProxyPort proxyPort = (ProxyPort) o;

            if (port != proxyPort.port) return false;
            if (!proxyInfo.equals(proxyPort.proxyInfo)) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = proxyInfo.hashCode();
            result = 31 * result + port;
            return result;
        }

        @Override
        public String toString() {
            return proxyInfo.getHostname() + ":" + port;
        }
    }

    private final List<ProxyPort> reachableProxies = new ArrayList<>();
    private final List<ProxyPort> unreachableProxies = new ArrayList<>();

    public void reportUnreachableProxy(SfaAuthority.ProxyInfo proxyInfo, int port) {
        assert proxyInfo != null;
        if (proxyInfo != null)
            unreachableProxies.add(new ProxyPort(proxyInfo, port));
    }

    public void reportReachableProxy(SfaAuthority.ProxyInfo proxyInfo, int port) {
        reachableProxies.add(new ProxyPort(proxyInfo, port));
    }

    public void resetProxyReachabilityCache() {
        reachableProxies.clear();
        unreachableProxies.clear();
    }

    public JFedConnection.SshProxyInfo chooseUserProxy(List<SfaAuthority.ProxyInfo> userAuthProxyInfo, GeniUser user, SshKeyInfo userSshKeyInfo) {
        if (reachableProxies.isEmpty() && unreachableProxies.isEmpty()) {
            //if no info available, automatically test
            waitForTryProxyTask(user);
        }

        final String username = user.getUserUrn().getEncodedResourceName();

        if (!reachableProxies.isEmpty()) {
            ProxyPort pp = reachableProxies.get(0);
            assert pp != null;
            LOG.debug("chooseUserProxy return reachableProxies 0 of " + reachableProxies.size() + ": " + pp + "  -> reachableProxies=" + reachableProxies + " unreachableProxies=" + unreachableProxies);
            return new JFedConnection.SshProxyInfo(
                    pp.getHostname(), pp.getPort(),
                    username /*user's username is always the username used at the authority proxy*/,
                    userSshKeyInfo /* authority proxy always uses user login file keys */,
                    pp.getProxyInfo().getHostKey(),
                    true);
        }

        for (SfaAuthority.ProxyInfo authProxyInfo : userAuthProxyInfo) {
            assert authProxyInfo.getProxyType().equals("jFed SSH Proxy");

            VlanRange portRange = new VlanRange(authProxyInfo.getPortRange());
            LOG.debug("chooseUserProxy is looking at proxy with ports=" + portRange + " while unreachableProxies=" + unreachableProxies);
            ProxyPort chosenPort = null;
            for (Integer port: portRange) {
                ProxyPort curPort = new ProxyPort(authProxyInfo, port);
                if (!unreachableProxies.contains(curPort)) {
                    chosenPort = curPort;
                    break;
                }
            }

            if (chosenPort == null) {
                LOG.debug("Proxy at " + authProxyInfo.getHostname() + " has no reachable ports, skipping.");
                continue;
            } else
                LOG.debug("chooseUserProxy search resulted in port=" + chosenPort.getPort());

            return new JFedConnection.SshProxyInfo(
                    authProxyInfo.getHostname(), chosenPort.getPort(),
                    username /*user's username is always the username used at the authority proxy*/,
                    userSshKeyInfo /* authority proxy always uses user login file keys */,
                    authProxyInfo.getHostKey(),
                    true);
        }
        return null;
    }

    public boolean isAnyProxyReachable() {
        return !reachableProxies.isEmpty();
    }

    public List<ProxyPort> getReachableProxies() {
        return Collections.unmodifiableList(reachableProxies);
    }

    public void setNeedProxyForJFed(boolean needProxyForJFed) {
        this.needProxyForJFed = needProxyForJFed;
        updateProxySettings();
    }

    public void setNeedProxyForSsh(boolean needProxyForSsh) {
        this.needProxyForSsh = needProxyForSsh;
        updateProxySettings();
    }


    public boolean isProxyEnabledForJfed() {
        String proxyUsage = jFedPreferences.getString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_JFED);
        if (proxyUsage.equalsIgnoreCase(NEVER_USE_PROXY))
            return false;
        if (proxyUsage.equalsIgnoreCase(AUTO_USE_PROXY))
            return needProxyForJFed;
        if (proxyUsage.equalsIgnoreCase(ALWAYS_USE_PROXY))
            return true;
        throw new RuntimeException("illegal value: proxyUsage=" + proxyUsage + "");
    }

    public boolean isProxyEnabledForSsh() {
        String proxyUsage = jFedPreferences.getString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH);
        if (proxyUsage.equalsIgnoreCase(NEVER_USE_PROXY))
            return false;
        if (proxyUsage.equalsIgnoreCase(AUTO_USE_PROXY))
            return needProxyForJFed;
        if (proxyUsage.equalsIgnoreCase(ALWAYS_USE_PROXY))
            return true;
        throw new RuntimeException("illegal value: proxyUsage=" + proxyUsage + "");
    }

    public JFedConnection.ProxyInfo getSshProxySettings(List<SfaAuthority.ProxyInfo> userAuthProxyInfo, GeniUser user, SshKeyInfo userSshKeyInfo) {
        return getProxySettings(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH, userAuthProxyInfo, user, userSshKeyInfo);
    }

    public JFedConnection.ProxyInfo getProxySettings(JFedPreferences.PreferenceKey preferencesKey, List<SfaAuthority.ProxyInfo> userAuthProxyInfo, GeniUser user, SshKeyInfo userSshKeyInfo) {
        String proxyUsage = jFedPreferences.getString(preferencesKey);
        if (proxyUsage.equalsIgnoreCase(NEVER_USE_PROXY)) {
            return null;
        }
        if (proxyUsage.equalsIgnoreCase(AUTO_USE_PROXY) && preferencesKey == GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH && !needProxyForSsh) {
            return null;
        }
        if (proxyUsage.equalsIgnoreCase(AUTO_USE_PROXY) && preferencesKey == GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_JFED && !needProxyForJFed) {
            return null;
        }
        if (proxyUsage.equalsIgnoreCase(ALWAYS_USE_PROXY) || (proxyUsage.equalsIgnoreCase(AUTO_USE_PROXY) && needProxyForSsh)) {
            if (jFedPreferences.getBoolean(GuiPreferenceKey.PREF_SSHPROXY_OVERRIDE)) {
                //set proxy info from preferences
                String hostname = jFedPreferences.getString(GuiPreferenceKey.PREF_SSHPROXY_HOSTNAME);
                int port = jFedPreferences.getInt(GuiPreferenceKey.PREF_SSHPROXY_PORT);
                String usernameAtProxy = jFedPreferences.getString(GuiPreferenceKey.PREF_SSHPROXY_USERNAME);
                SshKeyInfo sshKeyInfo = jFedPreferences.getOverriddenProxySshKeyInfo();
                String hostKey = jFedPreferences.getString(GuiPreferenceKey.PREF_SSHPROXY_HOSTKEY);

                JFedConnection.ProxyInfo prefProxyInfo =
                        new JFedConnection.SshProxyInfo(hostname, port, usernameAtProxy, sshKeyInfo, hostKey, false);
                return prefProxyInfo;
            } else {
                //find user proxy info (always use user login PEM keys, even if user has specified custom key)
                JFedConnection.SshProxyInfo userProxy = chooseUserProxy(userAuthProxyInfo, user, new UserSshKeyInfo(user));
                return userProxy;
            }
        }
        throw new RuntimeException("Unknown proxy setting for PREF_SSHPROXY_USE_FOR_SSH: \"" + proxyUsage + "\"");
    }

    /**
     * If needed, changes the settings of the connectionPool to reflect the new proxy settings.
     * This means this updates the proxy for new jFed's connections.
     * It does not change SSH settings. SSH proxy settings are requested directly by the ExternalSshTerminal object each time a terminal is started using getSshProxySettings().
     */
    public void updateProxySettings() {
        LOG.trace("ProxyPreferencesManager.updateProxySettings()");

        if (!userLoginModelManager.isUserLoggedIn()) return;
        GeniUser currentUser = userLoginModelManager.getLoggedInGeniUser();

        SfaAuthority userAuth = currentUser.getUserAuthority();

        JFedConnection.ProxyInfo usedProxy = getProxySettings(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_JFED,
                userAuth.getProxies(), currentUser, new UserSshKeyInfo(currentUser));

        if (prevUser == currentUser &&
                usedProxy != null && prevProxy != null && prevProxy.equals(usedProxy)
                && needProxyForJFed == prevNeedProxyForJFed)
            return; //no need to update, everything is the same

        LOG.trace("ProxyPreferencesManager.updateProxySettings() -> settings changed, will update.");

        prevUser = currentUser;
        prevNeedProxyForJFed = needProxyForJFed;
        prevProxy = usedProxy;

        //Activate proxy if present
        if (usedProxy != null) {
            LOG.info("SSH Proxy is activated for jFed calls, and there is a reachable SSH Proxy: " + usedProxy.getHostname() + ":" + usedProxy.getPort());
            connectionPool.setDefaultProxy(usedProxy);
            return;
        }

        LOG.info("No SSH Proxy for jFed calls");
        connectionPool.setDefaultProxy(null);
    }


    /**
     * check user authority proxy connectivity.
     */
//    public void tryProxy(GeniUser user) {
//        tryProxy(user.getUserAuthority().getProxies());
//    }
    private void waitForTask(Task task) {
        Thread taskThread = new Thread(task);
        taskThread.setDaemon(true);
        taskThread.setName("TryProxyThread");
        taskThread.start();

        try {
            taskThread.join();
        } catch (InterruptedException e) {
            LOG.error("TryProxyThread Task aborted", e);
        }
    }

    public void waitForTryProxyTask(GeniUser user) {
        if (user.getUserAuthority().getProxies() == null || user.getUserAuthority().getProxies().isEmpty())
            return;
        waitForTask(new TryProxyTask(user.getUserAuthority().getProxies()));
    }

    public void waitForTryProxyTask(SfaAuthority authority) {
        if (authority.getProxies() == null || authority.getProxies().isEmpty())
            return;
        waitForTask(new TryProxyTask(authority.getProxies()));
    }

    public Task<Boolean> getTryProxyTask(GeniUser user) {
        return new TryProxyTask(user.getUserAuthority().getProxies());
    }

    public TryProxyTask getTryProxyTask(SfaAuthority authority) {
        return new TryProxyTask(authority.getProxies());
    }


//    /**
//     * check user authority proxy connectivity.
//     */
//    public void tryProxy(SfaAuthority authority) {
//        tryProxy(authority.getProxies());
//    }

//    /**
//     * check user authority proxy connectivity.
//     */
//    public void tryProxy(List<SfaAuthority.ProxyInfo> userAuthProxyInfos) {
//        for (SfaAuthority.ProxyInfo proxyInfo : userAuthProxyInfos) {
//            assert proxyInfo.getProxyType().equals("jFed SSH Proxy");
//            tryProxy(proxyInfo);
//        }
//    }

    private static final int MAX_PORTS_TESTED = 10;

    /**
     * check user authority proxy connectivity.
     */
    public void tryProxy(SfaAuthority.ProxyInfo userAuthProxyInfo) {
        assert userAuthProxyInfo != null;
        int portsTested = 0;

        VlanRange portRange = new VlanRange(userAuthProxyInfo.getPortRange());
        for (Integer port: portRange) {
            if (testConnectivity(userAuthProxyInfo.getHostname(), port)) {
                reportReachableProxy(userAuthProxyInfo, port);
            } else {
                reportUnreachableProxy(userAuthProxyInfo, port);
            }

            //test maximum 10 ports for the same proxy
            if (portsTested++ >= MAX_PORTS_TESTED) return;

        }
    }

    public class TryProxyTask extends Task<Boolean> {
        private final List<SfaAuthority.ProxyInfo> userAuthProxyInfos;

        private TryProxyTask(List<SfaAuthority.ProxyInfo> userAuthProxyInfos) {
            this.userAuthProxyInfos = userAuthProxyInfos;
        }

        private boolean finished = false;
        private boolean success = false;
        public boolean isSuccess() { return success; }
        public boolean isFinished() { return finished; }

        @Override
        protected Boolean call() throws Exception {
            success = false;
            int portsTested = 0;

            updateMessage("Initializing proxy test");

            //count the total number of ports to be tested
            int totalPorts = 0;

            List<VlanRange> rangeHelpers = new ArrayList<>();
            for (SfaAuthority.ProxyInfo pi : userAuthProxyInfos) {
                VlanRange portRange = new VlanRange(pi.getPortRange());
                totalPorts += Math.min(portRange.size(), MAX_PORTS_TESTED);
                rangeHelpers.add(portRange);
            }

            resetProxyReachabilityCache();
            int totalPortsTested = 0;

            for (int i = 0; i < userAuthProxyInfos.size(); i++) {
                SfaAuthority.ProxyInfo userAuthProxyInfo = userAuthProxyInfos.get(i);
                assert userAuthProxyInfo != null;
                VlanRange portRange = rangeHelpers.get(i);

                assert userAuthProxyInfo.getProxyType().equals("jFed SSH Proxy");

                for (Integer port: portRange) {
                    //test maximum 10 ports for the same proxy
                    if (portsTested >= MAX_PORTS_TESTED)
                        break;

                    LOG.info("Testing proxy " + userAuthProxyInfo.getHostname() + ":" + port);
                    updateMessage(String.format("Testing proxy %s:%d", userAuthProxyInfo.getHostname(), port));

                    if (testConnectivity(userAuthProxyInfo.getHostname(), port)) {
                        LOG.info("-> proxy reachable " + userAuthProxyInfo.getHostname() + ":" + port);
                        reportReachableProxy(userAuthProxyInfo, port);
                        success = true;
                    } else {
                        reportUnreachableProxy(userAuthProxyInfo, port);
                        LOG.info("-> proxy unreachable " + userAuthProxyInfo.getHostname() + ":" + port);
                    }

                    totalPortsTested++;
                    updateProgress(totalPortsTested, totalPorts);

                    portsTested++;
                }

            }
            updateMessage("Finished testing proxies.");
            finished = true;
            return success;
        }
    }

    private static final int CONNECT_TIMEOUT = 5 * 1000; //5 seconds

    public boolean testConnectivity(String hostname, int port) {
        SocketAddress socketAddress = new InetSocketAddress(hostname, port);
        Socket socket = new Socket();
        try {
            //try to connect
            socket.connect(socketAddress, CONNECT_TIMEOUT);

            if (socket.isConnected())
                return true;
            else
                return false;
        } catch (SocketTimeoutException e) {
            return false;
        } catch (IOException e) {
            return false;
        } finally {
            try {
                socket.close();
            } catch (IOException ignored) {
            }
        }
    }
}
