package be.iminds.ilabt.jfed.experimenter_gui.editor.actions;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.experimenter_gui.editor.EditableRspec;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityInfo;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import javafx.scene.control.Alert;
import javafx.scene.control.ButtonType;
import javafx.scene.control.ChoiceDialog;
import javafx.scene.paint.Color;
import javafx.stage.Window;
import org.controlsfx.dialog.CommandLinksDialog;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * User: twalcari
 * Date: 5/20/2015
 * Time: 11:05
 */
public class ProcessUnboundNodesAction extends EditorAction<Boolean> {
    private static final Logger LOG = LoggerFactory.getLogger(ProcessUnboundNodesAction.class);

    private final RequestRspecSource requestRspecSource;


    private final AuthorityList authorityList;
    private final JFedConfiguration conf;

    ProcessUnboundNodesAction(EditableRspec editableRspec, Window parentWindow,
                              AuthorityList authorityList, JFedConfiguration conf) {
        super(editableRspec, parentWindow);
        this.authorityList = authorityList;
        this.conf = conf;

        this.requestRspecSource = editableRspec.getRequestRspecSource();
    }

    @Override
    public Boolean call() {
        FXModelRspec modelRspec = (FXModelRspec) requestRspecSource.getModelRspec(ModelRspecType.FX);
        if (modelRspec == null) {
            LOG.warn("Cannot process unbound nodes as modelRspec is null!");
            return false;
        }

        Collection<FXRspecNode> unboundNodes =
                modelRspec.getNodes().filtered(input1 -> input1.getComponentManagerId() == null);

        if (!unboundNodes.isEmpty()) {
            if (requestRspecSource.isLosingData(ModelRspecType.FX)) {
                LOG.warn("Cannot process unbound nodes as modelRspec is losing data!");

                Alert alert = new Alert(Alert.AlertType.WARNING);
                alert.setContentText("jFed detected unbound resources. These need to be bound to an authority before the experiment can be run." +
                        "jFed has detected that information will be lost when resolving this issue. Do you want to continue?");
                alert.setTitle("Unbound nodes detected");
                alert.setHeaderText("Unbound nodes detected");

                alert.getButtonTypes().setAll(ButtonType.YES, ButtonType.NO);

                if (alert.showAndWait().orElse(ButtonType.NO) == ButtonType.NO)
                    return false;

                LOG.info("User accepted data loss to resolve unbound nodes.");
            }

            CommandLinksDialog.CommandLinksButtonType resolveNowButtonType =
                    new CommandLinksDialog.CommandLinksButtonType("Assign these nodes now",
                            "Let me choose the testbed to assign the unbound nodes to.", true);


            CommandLinksDialog.CommandLinksButtonType resolveLaterButtonType =
                    new CommandLinksDialog.CommandLinksButtonType("Resolve later",
                            "I want to resolve this issue manually, or want to edit the unbound RSpec", false);

            CommandLinksDialog dialog = new CommandLinksDialog(resolveNowButtonType, resolveLaterButtonType);
            dialog.initOwner(parentWindow);
            dialog.setTitle("Unbound nodes detected");
            dialog.setHeaderText("Unbound nodes detected");
            dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.PUZZLE_PIECE, Color.CHOCOLATE));
            dialog.setContentText(String.format("This RSpec contains %d unbound nodes. " +
                            "These nodes need to be assigned to a testbed before starting the experiment.",
                    unboundNodes.size()));

            Optional<ButtonType> result = dialog.showAndWait();

            if (!result.isPresent() || result.get() != resolveNowButtonType.getButtonType())
                return false;

            List<AuthorityInfo> bindableAuthorities = new ArrayList<>();

            for (JFedConfiguration.TestbedDescription testbedDescription : conf.getTestbedDescriptions()) {
                if (testbedDescription.isShowAsBindable()) {
                    assert (authorityList.getByUrnExact(testbedDescription.getUrn()) != null) :
                            testbedDescription.getUrn() + " cannot be found in authorityList";

                    bindableAuthorities.add(authorityList.getByUrnExact(testbedDescription.getUrn()));
                }
            }

            Collections.sort(bindableAuthorities, Comparator.comparing(auth -> auth.getName().toLowerCase()));


            ChoiceDialog<AuthorityInfo> unboundNodesDialog = new ChoiceDialog<>();
            unboundNodesDialog.initOwner(parentWindow);
            unboundNodesDialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.PUZZLE_PIECE, Color.CHOCOLATE));
            unboundNodesDialog.setHeaderText("Unbound nodes detected");
            unboundNodesDialog.setContentText(String.format("Please select the testbed to which the %d unbound nodes must be assigned:",
                    unboundNodes.size()));
            unboundNodesDialog.getItems().setAll(bindableAuthorities);


            Optional<AuthorityInfo> chosenAuthorityInfo = unboundNodesDialog.showAndWait();


            if (chosenAuthorityInfo.isPresent()) {
                for (RspecNode rspecNode : unboundNodes) {
                    rspecNode.setComponentManagerId(chosenAuthorityInfo.get().getSfaAuthority().getUrn());
                }

                editableRspec.setRequestRspecSource(new RequestRspecSource(modelRspec));
                return true;
            }

            return false;
        } else
            return true;

    }


}
