package be.iminds.ilabt.jfed.experimenter_gui.editor.ontology.webmap;


import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.application.Platform;
import javafx.concurrent.Worker.State;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.Node;
import javafx.scene.control.*;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Priority;
import javafx.scene.layout.Region;
import javafx.scene.web.WebEngine;
import javafx.scene.web.WebView;
import javafx.util.Duration;
import netscape.javascript.JSObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * @author Jasper Potts
 */
public class ClusteringWebMap extends BorderPane {
    private static final Logger LOG = LoggerFactory.getLogger(ClusteringWebMap.class);

    private Timeline locationUpdateTimeline;

    private boolean updatingRectangle = false;
    private final WebView webView;
    final WebEngine webEngine;

    boolean initialized = false;

    public ClusteringWebMap() {
        // create web engine and view
        webView = new WebView();
        webEngine = webView.getEngine();
        webEngine.getLoadWorker().stateProperty().addListener(
                (ov, oldState, newState) -> {
                    if (newState == State.SUCCEEDED) {
                        JSObject win = (JSObject) webEngine.executeScript("window");
                        win.setMember("app", new JavaToMapBindingApp());

                        webEngine.executeScript("initialize()");

                        initialized = true;
                    }
                }
        );

        webEngine.load(getClass().getResource("clusteringgooglemap.html").toString());

        // create map type buttons
        final ToggleGroup mapTypeGroup = new ToggleGroup();
        final ToggleButton road = new ToggleButton("Road");
        road.setSelected(true);
        road.setToggleGroup(mapTypeGroup);
        final ToggleButton satellite = new ToggleButton("Satellite");
        satellite.setToggleGroup(mapTypeGroup);
        final ToggleButton hybrid = new ToggleButton("Hybrid");
        hybrid.setToggleGroup(mapTypeGroup);
        final ToggleButton terrain = new ToggleButton("Terrain");
        terrain.setToggleGroup(mapTypeGroup);
        mapTypeGroup.selectedToggleProperty().addListener((observableValue, toggle, toggle1) -> {
            if (road.isSelected()) {
                webEngine.executeScript("document.setMapTypeRoad()");
            } else if (satellite.isSelected()) {
                webEngine.executeScript("document.setMapTypeSatellite()");
            } else if (hybrid.isSelected()) {
                webEngine.executeScript("document.setMapTypeHybrid()");
            } else if (terrain.isSelected()) {
                webEngine.executeScript("document.setMapTypeTerrain()");
            }
        });

        webEngine.setOnAlert(text ->
                new Alert(Alert.AlertType.WARNING, text.toString()).show());


        // add search
        final TextField searchBox = new TextField();
        //searchBox.setColumns(12);
        searchBox.setPromptText("Search");
        searchBox.textProperty().addListener((observableValue, s, s1) -> {
            // delay location updates to we don't go too fast file typing
            if (locationUpdateTimeline != null) {
                locationUpdateTimeline.stop();
            }

            final Duration initialDelay = new Duration(400);
            final EventHandler<ActionEvent> performSearch = new EventHandler<ActionEvent>() {
                @Override
                public void handle(ActionEvent actionEvent) {
                    webEngine.executeScript("document.goToLocation(\"" + searchBox.getText() + "\")");
                }
            };

            locationUpdateTimeline = new Timeline(new KeyFrame(initialDelay, performSearch));
            locationUpdateTimeline.play();
        });

        Button zoomIn = new Button("Zoom In");
        zoomIn.setOnAction(actionEvent -> webEngine.executeScript("document.zoomIn()"));
        Button zoomOut = new Button("Zoom Out");
        zoomOut.setOnAction(actionEvent -> webEngine.executeScript("document.zoomOut()"));

        //workaround voor een bug die soms voorkomt (bug tussen google maps api en javafx 2.2)
        webView.setOnDragDetected(t -> {
            // intentionally do nothing
        });

        // create toolbar
        ToolBar toolBar = new ToolBar();
        toolBar.getStyleClass().add("map-toolbar");
        toolBar.getItems().addAll(
                road, satellite, hybrid, terrain,
                //                createSpacer(),
                //                google, yahoo, bing,
                createSpacer(),
                new Label("Location:"), searchBox, zoomIn, zoomOut);
        // create root
        getStyleClass().add("map");
        getStylesheets().add(getClass().getResource("WebMap.css").toString());
        setCenter(webView);
        setTop(toolBar);
    }

    private double roundCoordinate(double coord) {
        return Math.round(coord * 10_000) / 10_000.0;
    }

    @SuppressWarnings("unused") //these functions are called from Javascript!
    public class JavaToMapBindingApp {
        public void onMarkerClicked(String id){
            listeners.forEach(listener -> listener.onMarkerClicked(id));
        }

        public void onBoundsChanged(double minLat, double maxLat, double minLon, double maxLon){
            listeners.forEach(listener -> listener.onBoundsChanged(minLat, maxLat, minLon, maxLon));
        }

        public void log(String message) {
            LOG.info(message);
        }
    }

    private Node createSpacer() {
        Region spacer = new Region();
        HBox.setHgrow(spacer, Priority.ALWAYS);
        return spacer;
    }

    public void addMarker(String id, final double lat, final double lon) {
        Platform.runLater(() -> webEngine.executeScript("document.addMarker('"+id+"'," + lat + ", " + lon + ")"));
    }

    public void removeMarker(String id){
        Platform.runLater(() -> webEngine.executeScript("document.removeMarker('"+id+"')"));
    }

    public void clearMarkers(){
        Platform.runLater(() -> webEngine.executeScript("document.clearMarkers()"));
    }


    static { // use system proxy settings when standalone application    
        System.setProperty("java.net.useSystemProxies", "true");
    }

    public interface ClusteringWebMapListener {
        void onMarkerClicked(String id);
        void onBoundsChanged(double minLat, double maxLat, double minLon, double maxLon);
    }

    private final List<ClusteringWebMapListener> listeners = new ArrayList<>();

    public void addListener(ClusteringWebMapListener listener){
        listeners.add(listener);
    }


}
