package be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl;

import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.api.OCCI;
import be.iminds.ilabt.jfed.lowlevel.api.RestApi;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.HttpConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.ExtraInfoCallback;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.IOException;
import java.io.StringReader;
import java.net.URL;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * OcciToAggregateManagerWrapper
 * <p/>
 * OCCI doesn't provide a real AM, but we can wrap it (with limited functionality)...
 */
public class OcciToAggregateManagerWrapper extends AggregateManagerWrapper {
    public static final String bonfireUrnPart = "bonfire-project.eu";
    public static final GeniUrn bonfireUrn = GeniUrn.createGeniUrnFromEncodedParts(bonfireUrnPart, "authority", "cm");

    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(OcciToAggregateManagerWrapper.class);
    private static Map<String, GeniUrn> experimentIdToSliceUrn = new HashMap<String, GeniUrn>();
    static private Map<GeniUrn, ExperimentInfo> sliceUrnToExperimentInfo = new HashMap<GeniUrn, ExperimentInfo>();
    private OCCI occi;

    protected OcciToAggregateManagerWrapper(Logger logger, GeniUserProvider geniUserProvider,
                                            JFedConnectionProvider connectionProvider, SfaAuthority amAuthority, JFedPreferences jFedPreferences) {
        super(logger, geniUserProvider, connectionProvider, jFedPreferences, amAuthority);

        occi = new OCCI(logger, jFedPreferences);
    }

    //    public static GeniUrn mapExperimentInfoToSliceUrn(String experimentName) {
//        return new GeniUrn(bonfireUrnPart, "slice", experimentName);
//    }
    public static GeniUrn mapExperimentInfoToSliverUrn(SfaAuthority amAuthority, String experimentName) {
        return GeniUrn.createGeniUrnFromEncodedParts(amAuthority.getNameForUrn(), "sliver", experimentName);
    }

    public static GeniUrn mapSliceUrnToSliverUrn(SfaAuthority amAuthority, GeniUrn sliceUrn) {
        return GeniUrn.createGeniUrnFromEncodedParts(amAuthority.getNameForUrn(), "sliver", sliceUrn.getEncodedResourceName());
    }

    /**
     * HACK this is a shortcut, because SfaModelICCIListener has no way to know the slice URN, but it needs it...
     */
    public static synchronized GeniUrn mapExperimentIdToSliceUrn(String experimentId) {
        GeniUrn sliceUrn = experimentIdToSliceUrn.get(experimentId);
        assert sliceUrn != null : "No slice URN known for bonfire experiment ID: \"" + experimentId + "\"";
        return sliceUrn;
    }

    public static List<LocationInfo> parseLocationsXml(String xml) {

/* example
<?xml version="1.0" encoding="UTF-8"?>
<collection xmlns="http://api.bonfire-project.eu/doc/schemas/occi" href="/locations">
  <items offset="0" total="10">
<location href="/locations/autobahn">
  <name>autobahn</name>
  <url>http://172.18.9.5:8080/autobahn/uap</url>
</location>
<location href="/locations/be-ibbt">
  <name>be-ibbt</name>
  <url>https://bonfire2.test.atlantis.ugent.be</url>
</location>
* */
        try {
            List<LocationInfo> locations = new ArrayList<LocationInfo>();

            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            InputSource is = new InputSource(new StringReader(xml));
            Document doc = db.parse(is);

            Element rootEl = doc.getDocumentElement();

            String rootElName = rootEl.getTagName();
            if (!rootElName.equalsIgnoreCase("collection")) {
                throw new RuntimeException("Not a collection (root element tag name=" + rootElName + "): " + xml);
            }

            NodeList itemsList = rootEl.getElementsByTagName("items");
            assert itemsList.getLength() > 0;
            Node itemsN = itemsList.item(0);
            assert itemsN instanceof Element;
            Element itemsEl = (Element) itemsN;

            NodeList locationList = itemsEl.getElementsByTagName("location");
            for (int i = 0; i < locationList.getLength(); i++) {
                Node locNode = locationList.item(i);
                assert locNode instanceof Element;
                Element locEl = (Element) locNode;

                String href = locEl.getAttribute("href");

                NodeList nameList = locEl.getElementsByTagName("name");
                assert nameList.getLength() == 1;
                Node nameN = nameList.item(0);
                assert nameN instanceof Element;
                Element nameEl = (Element) nameN;

                NodeList urlList = locEl.getElementsByTagName("url");
                assert urlList.getLength() == 1;
                Node urlN = urlList.item(0);
                assert urlN instanceof Element;
                Element urlEl = (Element) urlN;

                LocationInfo locationInfo = new LocationInfo(href, nameEl.getTextContent(), urlEl.getTextContent());
                locations.add(locationInfo);
            }

            return locations;
        } catch (ParserConfigurationException e) {
            throw new RuntimeException("ParserConfigurationException: " + e.getMessage(), e);
        } catch (SAXException e) {
            throw new RuntimeException("SAXException: " + e.getMessage(), e);
        } catch (IOException e) {
            throw new RuntimeException("IOException: " + e.getMessage(), e);
        }
    }

    public static String fakeAdvertisementRspec(SfaAuthority amAuthority, List<LocationInfo> locationInfos) {
        assert locationInfos != null;

        StringBuilder adRspecBuilder = new StringBuilder();
        adRspecBuilder.append("<rspec type=\"advertisement\" xmlns=\"http://www.geni.net/resources/rspec/3\"" +
                "    xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/ad.xsd \">\n");
        for (LocationInfo locationInfo : locationInfos) {
            String site = locationInfo.getName();
            String siteUrnPart = site.replaceAll("[^a-zA-Z0-9-]", "");
            adRspecBuilder.append("    <node\n" + "        component_id=\"urn:publicid:IDN+" + amAuthority.getNameForUrn() + "+node+").
                    append(siteUrnPart).
                    append("\"\n").
                    append("        component_manager_id=\"").
                    append(bonfireUrn.toString()).
                    append("\"\n").
                    append("        component_name=\"").
                    append(site).
                    append("\" exclusive=\"false\">\n").
                    append("        <sliver_type name=\"lite\"/>\n").
                    append("        <sliver_type name=\"small\"/>\n").
                    append("    </node>\n");
        }
        adRspecBuilder.append("</rspec>");
        final String adRspec = adRspecBuilder.toString();
        return adRspec;
    }

    /**
     * @return ExperimentInfo with experimentId and experimentName filled in
     */
    public static ExperimentInfo parseExperiment(SfaAuthority auth, GeniUrn sliceUrn, String xml) {
        /*
        <experiment href="/experiments/336" xmlns="http://api.bonfire-project.eu/doc/schemas/occi">
          <id>336</id>
          ...
        * */
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            InputSource is = new InputSource(new StringReader(xml));
            Document doc = db.parse(is);

            Element rootEl = doc.getDocumentElement();

            String rootElName = rootEl.getTagName();
            if (!rootElName.equalsIgnoreCase("experiment")) {
                throw new RuntimeException("Not an experiment (root element tag name=" + rootElName + "): " + xml);
            }

            NodeList idList = rootEl.getElementsByTagName("id");
            assert idList.getLength() > 0;
            Node idN = idList.item(0);
            assert idN instanceof Element;
            Element idEl = (Element) idN;
            String id = idEl.getTextContent();

            NodeList nameList = rootEl.getElementsByTagName("name");
            assert nameList.getLength() > 0;
            Node nameN = nameList.item(0);
            assert nameN instanceof Element;
            Element nameEl = (Element) nameN;
            String name = nameEl.getTextContent();

            GeniUrn sliverUrn = mapExperimentInfoToSliverUrn(auth, name);
            assert sliceUrn == null || sliverUrn.equals(mapSliceUrnToSliverUrn(auth, sliceUrn));

            ExperimentInfo res = new ExperimentInfo(sliceUrn, sliverUrn, name);
            res.experimentId = idEl.getTextContent();
            return res;
        } catch (ParserConfigurationException e) {
            throw new RuntimeException("ParserConfigurationException: " + e.getMessage(), e);
        } catch (SAXException e) {
            throw new RuntimeException("SAXException: " + e.getMessage(), e);
        } catch (IOException e) {
            throw new RuntimeException("IOException: " + e.getMessage(), e);
        }
    }

    private static String fixedPublicNetworkIdForLocation(String locationName) {
        //TODO get list using OCCI and let user choose using RSpec somehow
        if (locationName.equals("uk-epcc")) return "517";
        if (locationName.equals("be-ibbt")) return null;
        if (locationName.equals("de-hlrs")) return null;
        if (locationName.equals("es-wellness")) return null;
        if (locationName.equals("fr-inria")) return "27";
        if (locationName.equals("pl-psnc")) return null;
        if (locationName.equals("uk-hplabs")) return "subnet-2-2";
        return null;
    }

    private static String fixedWanNetworkIdForLocation(String locationName) {
        //TODO get list using OCCI and let user choose using RSpec somehow
        if (locationName.equals("uk-epcc")) return "105";
        if (locationName.equals("be-ibbt")) return "1"; //also 2561 2563
        if (locationName.equals("de-hlrs")) return "90";
        if (locationName.equals("es-wellness")) return "8d6b0daa-858b-41e7-b72e-310ac8619b50";
        if (locationName.equals("fr-inria")) return "53";
        if (locationName.equals("pl-psnc")) return "0";
        if (locationName.equals("uk-hplabs")) return "subnet-2-3";
        return null;
    }

    private static String fixedStorageIdForLocation(String locationName) {
        //TODO get list using OCCI and let user choose using RSpec somehow
        if (locationName.equals("be-ibbt")) return "14";
        if (locationName.equals("de-hlrs")) return "3516";
        if (locationName.equals("es-wellness")) return "vappTemplate-394cf26e-85a9-4180-80fd-88e8f6e47dc1";
        if (locationName.equals("fr-inria")) return "3904";
        if (locationName.equals("pl-psnc")) return "1067";
        if (locationName.equals("uk-epcc")) return "4527";
        if (locationName.equals("uk-hplabs")) return "vol-32108796";
        return null;
    }

    public static List<ComputeInfo> parseExperimentComputes(String experimentId, String xml) {
        /*example:
<collection xmlns="http://api.bonfire-project.eu/doc/schemas/occi" href="/experiments/336/computes">
  <items offset="0" total="1">
<compute href="/locations/fr-inria/computes/426" name="BonFIRE-monitor-experiment336"/>
  </items>
  <link href="/experiments/336" rel="parent" type="application/vnd.bonfire+xml"/>
</collection>
        * */

        List<ComputeInfo> res = new ArrayList<ComputeInfo>();
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            InputSource is = new InputSource(new StringReader(xml));
            Document doc = db.parse(is);

            Element rootEl = doc.getDocumentElement();

            String rootElName = rootEl.getTagName();
            if (!rootElName.equalsIgnoreCase("collection")) {
                throw new RuntimeException("Not an collection (root element tag name=" + rootElName + "): " + xml);
            }

            NodeList itemsList = rootEl.getElementsByTagName("items");
            assert itemsList.getLength() > 0;
            Node itemN = itemsList.item(0);
            assert itemN instanceof Element;
            Element itemEl = (Element) itemN;

            NodeList computesList = itemEl.getChildNodes();
            for (int i = 0; i < computesList.getLength(); i++) {
                Node n = computesList.item(i);

                if (n instanceof Element) {
                    Element e = (Element) n;
                    String childName = e.getTagName();
                    ComputeInfo computeInfo = new ComputeInfo();
                    if (childName.equalsIgnoreCase("compute")) {
                        computeInfo.computeName = e.getAttribute("name");
                        String href = e.getAttribute("href");
                        Pattern pattern = Pattern.compile("/locations/([a-zA-Z0-9-_]*)/computes/([0-9]*)");
                        Matcher matcher = pattern.matcher(href);
                        if (matcher.find()) {
                            computeInfo.locationName = matcher.group(1);
                            computeInfo.computeId = matcher.group(2);
                        } else
                            LOG.warn("found no location in compute href=\"" + href + "\"");
                        computeInfo.experimentId = experimentId;
                        res.add(computeInfo);
                    }
                }
            }
        } catch (ParserConfigurationException e) {
            throw new RuntimeException("ParserConfigurationException: " + e.getMessage(), e);
        } catch (SAXException e) {
            throw new RuntimeException("SAXException: " + e.getMessage(), e);
        } catch (IOException e) {
            throw new RuntimeException("IOException: " + e.getMessage(), e);
        }
        return res;
    }

    public static List<StorageInfo> parseStorageCollection(String xml) {
        /*example
        *<collection xmlns="http://api.bonfire-project.eu/doc/schemas/occi">
          <items>
            <storage href="/locations/uk-epcc/storages/44" name="data"/>
            <storage href="/locations/uk-epcc/storages/47" name="data2"/>
            <storage href="/locations/uk-epcc/storages/74" name="whatever name"/>
            <storage href="/locations/uk-epcc/storages/91" name="Squeeze"/>
            <storage href="/locations/uk-epcc/storages/92" name="Squeeze 2G"/>
            <storage href="/locations/uk-epcc/storages/93" name="Zabbix Aggregator"/>
          </items>
        </collection>*/
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            InputSource is = new InputSource(new StringReader(xml));
            Document doc = db.parse(is);

            Element rootEl = doc.getDocumentElement();

            String rootElName = rootEl.getTagName();
            if (!rootElName.equalsIgnoreCase("collection")) {
                throw new RuntimeException("Not a collection (root element tag name=" + rootElName + "): " + xml);
            }

            NodeList itemsList = rootEl.getElementsByTagName("items");
            assert itemsList.getLength() > 0;
            Node itemsN = itemsList.item(0);
            assert itemsN instanceof Element;
            Element itemsEl = (Element) itemsN;

            List<StorageInfo> resStorages = new ArrayList<StorageInfo>();
            NodeList storagesList = itemsEl.getChildNodes();
            for (int i = 0; i < storagesList.getLength(); i++) {
                Node n = storagesList.item(i);

                if (n instanceof Element) {
                    Element e = (Element) n;
                    String href = e.getAttribute("href");
                    String name = e.getAttribute("name");

                    StorageInfo si = new StorageInfo(href, name);
                    resStorages.add(si);
                }
            }

            return resStorages;
        } catch (ParserConfigurationException e) {
            throw new RuntimeException("ParserConfigurationException: " + e.getMessage(), e);
        } catch (SAXException e) {
            throw new RuntimeException("SAXException: " + e.getMessage(), e);
        } catch (IOException e) {
            throw new RuntimeException("IOException: " + e.getMessage(), e);
        }
    }

    public static List<NetworkInfo> parseNetworkCollection(String xml) {
        /*example
        *<collection xmlns="http://api.bonfire-project.eu/doc/schemas/occi">
          <items>
            <network href="/locations/uk-epcc/networks/44" name="data"/>
            <network href="/locations/uk-epcc/networks/47" name="data2"/>
            <network href="/locations/uk-epcc/networks/74" name="whatever name"/>
            <network href="/locations/uk-epcc/networks/91" name="Squeeze"/>
            <network href="/locations/uk-epcc/networks/92" name="Squeeze 2G"/>
            <network href="/locations/uk-epcc/networks/93" name="Zabbix Aggregator"/>
          </items>
        </collection>*/
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            InputSource is = new InputSource(new StringReader(xml));
            Document doc = db.parse(is);

            Element rootEl = doc.getDocumentElement();

            String rootElName = rootEl.getTagName();
            if (!rootElName.equalsIgnoreCase("collection")) {
                throw new RuntimeException("Not a collection (root element tag name=" + rootElName + "): " + xml);
            }

            NodeList itemsList = rootEl.getElementsByTagName("items");
            assert itemsList.getLength() > 0;
            Node itemsN = itemsList.item(0);
            assert itemsN instanceof Element;
            Element itemsEl = (Element) itemsN;

            List<NetworkInfo> resNetworks = new ArrayList<NetworkInfo>();
            NodeList networksList = itemsEl.getChildNodes();
            for (int i = 0; i < networksList.getLength(); i++) {
                Node n = networksList.item(i);

                if (n instanceof Element) {
                    Element e = (Element) n;
                    String href = e.getAttribute("href");
                    String name = e.getAttribute("name");

                    NetworkInfo si = new NetworkInfo(href, name);
                    resNetworks.add(si);
                }
            }

            return resNetworks;
        } catch (ParserConfigurationException e) {
            throw new RuntimeException("ParserConfigurationException: " + e.getMessage(), e);
        } catch (SAXException e) {
            throw new RuntimeException("SAXException: " + e.getMessage(), e);
        } catch (IOException e) {
            throw new RuntimeException("IOException: " + e.getMessage(), e);
        }
    }

    /**
     * @return state
     */
    public static ComputeInfo parseCompute(String xml) {
        /*example
        * <compute xmlns="http://api.bonfire-project.eu/doc/schemas/occi" href="/locations/fr-inria/computes/427">
          <id>427</id>
          <cpu>1</cpu>
          <memory>1024</memory>
          <name>server-experiment#336</name>
          <instance_type>small</instance_type>
          <state>PENDING</state>
          <disk id="0">
            <storage href="/locations/fr-inria/storages/64" name="squeeze 2G 2" />
            <type>DISK</type>
            <target>xvda</target>
          </disk>
          <nic>
            <network href="/locations/fr-inria/networks/20" name="Public Network" />
            <ips>131.254.204.145</ips>
            <mac>02:00:83:fe:cc:91</mac>
          </nic>
          <context>
        <eth0_ip>129.215.175.149</eth0_ip>
          </context>
        </compute>*/
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            InputSource is = new InputSource(new StringReader(xml));
            Document doc = db.parse(is);

            Element rootEl = doc.getDocumentElement();

            String rootElName = rootEl.getTagName();
            if (!rootElName.equalsIgnoreCase("compute")) {
                throw new RuntimeException("Not a compute (root element tag name=" + rootElName + "): " + xml);
            }

            ComputeInfo computeInfo = new ComputeInfo();
            String href = rootEl.getAttribute("href");
            Pattern pattern = Pattern.compile("/locations/([a-zA-Z0-9-_]*)/computes/([0-9]*)");
            Matcher matcher = pattern.matcher(href);
            if (matcher.find()) {
                computeInfo.locationName = matcher.group(1);
                computeInfo.computeId = matcher.group(2);
            } else
                LOG.warn("found no location in compute href=\"" + href + "\"");

            NodeList computesList = rootEl.getChildNodes();
            for (int i = 0; i < computesList.getLength(); i++) {
                Node n = computesList.item(i);

                if (n instanceof Element) {
                    Element e = (Element) n;
                    String childName = e.getTagName();
                    if (childName.equalsIgnoreCase("state")) {
                        computeInfo.state = e.getTextContent();
                    }
                    if (childName.equalsIgnoreCase("id")) {
                        assert computeInfo.computeId == null || computeInfo.computeId.equals(e.getTextContent()) :
                                "compute id mismatch";
                        computeInfo.computeId = e.getTextContent();
                    }
                    if (childName.equalsIgnoreCase("name"))
                        computeInfo.computeName = e.getTextContent();

//                    if (childName.equalsIgnoreCase("context")) {
//                        NodeList contextNodesList = e.getChildNodes();
//                        for (int j = 0; j < contextNodesList.getLength(); j++) {
//                            Node subn = contextNodesList.item(j);
//                            if (subn instanceof Element) {
//                                Element sube = (Element) subn;
//                                String subChildName = sube.getTagName();
//                                if (subChildName.startsWith("eth") &&
//                                    subChildName.endsWith("_ip"))
//                                    computeInfo.ipsByNetworkId.put(?, sube.getTextContent());
//                            }
//                        }
//                    }
                    if (childName.equalsIgnoreCase("nic")) {
                        NodeList nicNodesList = e.getChildNodes();
                        String nicIp = null, networkId = null;
                        for (int j = 0; j < nicNodesList.getLength(); j++) {
                            Node subn = nicNodesList.item(j);
                            if (subn instanceof Element) {
                                Element sube = (Element) subn;
                                String subChildName = sube.getTagName();
                                if (subChildName.equals("network")) {
                                    String networkHref = sube.getAttribute("href");
                                    int lastSlash = networkHref.lastIndexOf('/');
                                    if (lastSlash > 0) {
                                        networkId = networkHref.substring(lastSlash + 1);
                                    }
                                }
                                if (subChildName.equals("ip")) {
                                    nicIp = sube.getTextContent();
                                }
                            }
                        }
                        if (nicIp != null && networkId != null) {
                            computeInfo.ipsByNetworkId.put(networkId, nicIp);
                            LOG.debug("Parsed IP in bonfire compute info: " + networkId + " -> " + nicIp);
                        }
                    }
                }
            }

            return computeInfo;
        } catch (ParserConfigurationException e) {
            throw new RuntimeException("ParserConfigurationException: " + e.getMessage(), e);
        } catch (SAXException e) {
            throw new RuntimeException("SAXException: " + e.getMessage(), e);
        } catch (IOException e) {
            throw new RuntimeException("IOException: " + e.getMessage(), e);
        }
    }

    public String mapSliceUrnToExperimentName(GeniUrn sliceUrn) {
        return sliceUrn.getEncodedResourceName();
    }

    public String mapSliverUrnToExperimentName(GeniUrn sliceUrn) {
        return sliceUrn.getEncodedResourceName();
    }

    protected HttpConnection getConnection() throws JFedException {
        JFedConnection res = getConnection(new ServerType(ServerType.GeniServerRole.OCCI, 1));
        if (res == null) {
            LOG.error("No OCCI connection can be created for auth with urls: "+amAuthority.getUrls());
            throw new RuntimeException("No connection can be created. Perhaps the authority \"" + amAuthority.getUrnString() + "\" does not have an OCCI interface, or no login details were given?");
        }
        return (HttpConnection) res;
    }

    @Override
    public Hashtable getVersion() throws JFedException {
        return null;
    }

    @Override
    public String listResources(List<AnyCredential> userCredentials, boolean available, String type, String version) throws JFedException {
        assert type.equals("geni");
        assert version.equals("3");

        HttpConnection con;
        try {
            con = getConnection();
        } catch (ExtraInfoCallback.NoBonfireAccountException e) {
            return "";
        }
        RestApi.RestReply<String> reply = occi.getLocations(con);

        if (!reply.getGeniResponseCode().isSuccess()) {
            throw new JFedException("OCCI getExperiments Call not successful");
        }

        List<LocationInfo> locations = parseLocationsXml(reply.getValue());
        LOG.debug("OcciToAggregateManagerWrapper.listResources: Received " + locations.size() + " locations from server: " + locations);
        String fakeAdvertisementRspec = fakeAdvertisementRspec(amAuthority, locations);
        LOG.debug("OcciToAggregateManagerWrapper.listResources: Created fake advertisement Rspec: " + fakeAdvertisementRspec);
        return fakeAdvertisementRspec;
    }

    @Override
    public AllocateReply allocate(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, Date startDate, Date endDate) throws JFedException {
        throw new UnsupportedOperationException();
    }

    @Override
    public String createSliver(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, List<UserSpec> users, Date expirationDate) throws JFedException {
        HttpConnection con;
        try {
            con = getConnection();
        } catch (ExtraInfoCallback.NoBonfireAccountException e) {
            return null;
        }

        //create bonfire experiment and generate sliver URN. link (static) sliver URN with slice URN and with bonfire experiment in this class.

        LOG.debug("OcciToAggregateManagerWrapper.createSliver sliceUrn=" + sliceUrn + " rspec=" + rspec);

        String groups = ExtraInfoCallback.getBonfireGroups();

        ExperimentInfo exInf = sliceUrnToExperimentInfo.get(sliceUrn);
        if (exInf != null && exInf.experimentId != null) {
            //check if experiment has compute nodes. If it has fail.
            RestApi.RestReply<String> reply = occi.getExperimentComputes(con, exInf.experimentId);

            //TODO check if there are computes...
            LOG.warn("TO IMPLEMENT: check if there are computes active already, fail if so");
        } else {
            exInf =
                    new ExperimentInfo(sliceUrn, mapExperimentInfoToSliverUrn(amAuthority, sliceUrn.getEncodedResourceName()), sliceUrn.getEncodedResourceName());
            LOG.debug("Linking generated experiment Urn (" + sliceUrn + ") with bonfire experiment info");
            sliceUrnToExperimentInfo.put(sliceUrn, exInf);
            exInf.experimentName = sliceUrn.getEncodedResourceName();

            //walltime = time in seconds that the sliver should live
            int walltime_s = (int) ((expirationDate.getTime() - new Date().getTime()) / 1000);

            String description = "experiment generated by jFed";

            RestApi.RestReply<String> reply =
                    occi.postExperimentWizard(con, groups, exInf.experimentName, walltime_s, description);

            if (!reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
                throw new JFedException("Failed to create experiment \"" + exInf.experimentName + "\"");

            //parse experiment ID
            exInf.experimentId = parseExperiment(amAuthority, sliceUrn, reply.getValue()).experimentId;
            experimentIdToSliceUrn.put(exInf.experimentId, sliceUrn); //register link between experiment ID and slice
            LOG.debug("mapped experiment id " + exInf.experimentId + " to slice urn " + sliceUrn);
        }

        assert exInf.experimentName != null;
        assert exInf.experimentId != null;

        StringBuilder fakeManifestRspecBuilder = new StringBuilder(200);
        fakeManifestRspecBuilder.append("<rspec " +
                "xmlns:client=\"http://www.protogeni.net/resources/rspec/ext/client/1\" " +
                "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " +
                "xmlns=\"http://www.geni.net/resources/rspec/3\" type=\"manifest\" " +
                "xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/manifest.xsd \" >\n");

        //create computes
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setNamespaceAware(true);
            DocumentBuilder db = dbf.newDocumentBuilder();
            InputSource is = new InputSource(new StringReader(rspec));
            Document doc = db.parse(is);

            Element rootEl = doc.getDocumentElement();

            String rootElName = rootEl.getTagName();
            if (!rootElName.equalsIgnoreCase("rspec")) {
                throw new RuntimeException("Not an rspec (root element tag name=" + rootElName + "): " + rspec);
            }

            NodeList childNodeList = rootEl.getChildNodes();
            for (int i = 0; i < childNodeList.getLength(); i++) {
                Node n = childNodeList.item(i);

                if (n instanceof Element) {
                    Element e = (Element) n;
                    String childName = e.getTagName();
                    if (childName.equalsIgnoreCase("node")) {
                        String componentId = e.getAttribute("component_id");
                        String componentManagerId = e.getAttribute("component_manager_id");
                        String clientId = e.getAttribute("client_id");

                        boolean usableRspecNode = true;
                        if (componentId == null || componentId.trim().isEmpty()) usableRspecNode = false;
                        if (componentManagerId == null || componentManagerId.trim().isEmpty()) usableRspecNode = false;
                        if (clientId == null || clientId.trim().isEmpty()) usableRspecNode = false;
                        if (!usableRspecNode)
                            LOG.warn("RSpec node is not usable for bonfire emulation: componentId=" + componentId +
                                    " clientId=" + clientId + " componentManagerId=" + componentManagerId);

                        if (usableRspecNode && componentManagerId.equals(bonfireUrn.toString())) {
                            ComputeInfo computeInfo = new ComputeInfo();
                            computeInfo.computeName = clientId;
                            GeniUrn locationUrn = new GeniUrn(componentId);
                            computeInfo.locationName = locationUrn.getEncodedResourceName();

                            //xsd schema:
//                            <?xml version="1.0"?>
//                            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"
//                                       targetNamespace="http://jfed.iminds.be/rspec/ext/jfed-bonfire/1"
//                                       xmlns="http://jfed.iminds.be/rspec/ext/jfed-bonfire/1">
//                                <xs:element name="bonfire_node_info">
//                                    <xs:complexType>//
//                                        <xs:sequence>
//                                            <xs:element name="network" minOccurs="0" maxOccurs="unbounded">
//                                                <xs:complexType>
//                                                    <xs:attribute name="href" type="xs:string" use="required"/>
//                                                </xs:complexType>
//                                            </xs:element>
//
//                                            <xs:element name="storage" minOccurs="0" maxOccurs="unbounded">
//                                                <xs:complexType>
//                                                    <xs:attribute name="href" type="xs:string" use="required"/>
//                                                </xs:complexType>
//                                            </xs:element>
//                                        </xs:sequence>
//                                        <xs:attribute name="add_public_network" type="xs:boolean" default="false"/>
//                                        <xs:attribute name="add_wan_network" type="xs:boolean" default="true"/>
//                                        <xs:attribute name="add_default_disk" type="xs:boolean" default="true"/>
//                                        <xs:attribute name="instance_type" type="xs:string" default="lite"/>
//                                    </xs:complexType>
//                                </xs:element>
//                            </xs:schema>
                            //example xml (to be added to rspec node):
//                            <bonfire_node_info add_public_network="true" add_wan_network="false" instance_type="lite">
//                                <network href="ADD NETWORK HREF HERE"/>
//                                <storage href="ADD DISK HREF HERE"/>
//                            </bonfire_node_info>

                            String instanceType = "lite";
                            boolean usedBonfireInfoNode = false;

                            //get info from jFed bonfire rspec extension
                            NodeList bonfireInfoNodes =
                                    e.getElementsByTagNameNS("http://jfed.iminds.be/rspec/ext/jfed-bonfire/1", "bonfire_node_info");
                            if (bonfireInfoNodes.getLength() > 0) {
                                Node bonfireInfoNode = bonfireInfoNodes.item(0);
                                assert bonfireInfoNode != null;

                                if (bonfireInfoNode instanceof Element) {
                                    usedBonfireInfoNode = true;

                                    Element bonfireInfoEl = (Element) bonfireInfoNode;
                                    assert bonfireInfoEl != null;

//                                    String addPublicNetAtt = bonfireInfoEl.getAttribute("add_public_network");
//                                    String addWanNetAtt = bonfireInfoEl.getAttribute("add_wan_network");
//                                    String addDefaultDiskAtt = bonfireInfoEl.getAttribute("add_default_disk");
                                    String instanceTypeAtt = bonfireInfoEl.getAttribute("instance_type");

//                                    if (publicNetId != null && addPublicNetAtt != null && TextUtil.stringToBoolean(addPublicNetAtt)) //default false
//                                        computeInfo.networkIds.add(publicNetId);
//                                    if (wanNetId != null && (addWanNetAtt == null || TextUtil.stringToBoolean(addWanNetAtt))) //default true
//                                        computeInfo.networkIds.add(wanNetId);
//                                    if (addDefaultDiskAtt == null || TextUtil.stringToBoolean(addDefaultDiskAtt)) //default true
//                                        computeInfo.storageIds.add(fixedStorageIdForLocation(computeInfo.locationName));
                                    if (instanceTypeAtt != null)
                                        instanceType = instanceTypeAtt;
                                    else {
                                        instanceType = "lite";
                                        LOG.warn("Parsed bonfire XML: No instancetype, using default: " + instanceType);
                                    }

                                    NodeList bonfireNetworkNodes =
                                            bonfireInfoEl.getElementsByTagNameNS("http://jfed.iminds.be/rspec/ext/jfed-bonfire/1", "network");
                                    for (int b = 0; b < bonfireNetworkNodes.getLength(); b++)
                                        if (bonfireNetworkNodes.item(b) instanceof Element) {
                                            Element bonfireNetworkNode = (Element) bonfireNetworkNodes.item(b);
                                            String bonfireNetworkNodeHrefAtt = bonfireNetworkNode.getAttribute("href");
                                            if (bonfireNetworkNodeHrefAtt != null) {
                                                String start =
                                                        "https://api.bonfire-project.eu/locations/" + computeInfo.locationName + "/networks/";
                                                if (bonfireNetworkNodeHrefAtt.startsWith(start)) {
                                                    String networkId =
                                                            bonfireNetworkNodeHrefAtt.substring(start.length());
                                                    computeInfo.networkIds.add(networkId);
                                                    LOG.debug("Parsed bonfire XML: added network with networkId=" + networkId);
                                                } else {
                                                    String start2 =
                                                            "/locations/" + computeInfo.locationName + "/networks/";
                                                    if (bonfireNetworkNodeHrefAtt.startsWith(start2)) {
                                                        String networkId =
                                                                bonfireNetworkNodeHrefAtt.substring(start2.length());
                                                        computeInfo.networkIds.add(networkId);
                                                        LOG.debug("Parsed bonfire XML: added network with networkId=" + networkId);
                                                    } else {
                                                        LOG.warn("Suspisious bonfire XML <network href=\"" + bonfireNetworkNodeHrefAtt + "\">. Is this a valid network? Will be tried anyway.");
                                                        computeInfo.networkIds.add(bonfireNetworkNodeHrefAtt);
                                                    }
                                                }
                                            } else
                                                LOG.error("Invalid bonfire XML <network> is missing href. Will be ignored.");
                                        } else
                                            LOG.error("Invalid bonfire XML <network> is not an Element but a " + bonfireNetworkNodes.item(b).getClass().getName() + ". Will be ignored.");

                                    NodeList bonfireStorageNodes =
                                            bonfireInfoEl.getElementsByTagNameNS("http://jfed.iminds.be/rspec/ext/jfed-bonfire/1", "storage");
                                    for (int b = 0; b < bonfireStorageNodes.getLength(); b++)
                                        if (bonfireStorageNodes.item(b) instanceof Element) {
                                            Element bonfireStorageNode = (Element) bonfireStorageNodes.item(b);
                                            String bonfireStorageNodeHrefAtt = bonfireStorageNode.getAttribute("href");
                                            if (bonfireStorageNodeHrefAtt != null) {
                                                String start =
                                                        "https://api.bonfire-project.eu/locations/" + computeInfo.locationName + "/storages/";
                                                if (bonfireStorageNodeHrefAtt.startsWith(start)) {
                                                    String storageId =
                                                            bonfireStorageNodeHrefAtt.substring(start.length());
                                                    computeInfo.storageIds.add(storageId);
                                                    LOG.debug("Parsed bonfire XML: added storage with storageId=" + storageId);
                                                } else {
                                                    String start2 =
                                                            "/locations/" + computeInfo.locationName + "/storages/";
                                                    if (bonfireStorageNodeHrefAtt.startsWith(start2)) {
                                                        String storageId =
                                                                bonfireStorageNodeHrefAtt.substring(start2.length());
                                                        computeInfo.storageIds.add(storageId);
                                                        LOG.debug("Parsed bonfire XML: added storage with storageId=" + storageId);
                                                    } else {
                                                        LOG.warn("Suspisious bonfire XML <storage href=\"" + bonfireStorageNodeHrefAtt + "\">. Is this a valid storage? Will be tried anyway.");
                                                        computeInfo.storageIds.add(bonfireStorageNodeHrefAtt);
                                                    }
                                                }
                                            } else
                                                LOG.error("Invalid bonfire XML <storage> is missing href. Will be ignored.");
                                        } else
                                            LOG.error("Invalid bonfire XML <storage> is not an Element but a " + bonfireStorageNodes.item(b).getClass().getName() + ". Will be ignored.");

                                    LOG.debug("Parsed bonfire XML: parsed instanceType=" + instanceType + ", " + bonfireNetworkNodes.getLength() + " networks and " + bonfireStorageNodes.getLength() + " storages.");
                                } else
                                    LOG.error("Invalid bonfire XML <jFedBonfire:bonfire_node_info> is not an Element but " +
                                            "a " + (bonfireInfoNode == null ? "null "
                                            : bonfireInfoNode.getClass().getName()) +
                                            ". Will be ignored.");
                            } else
                                LOG.warn("No bonfire node info, there are " +
                                        (bonfireInfoNodes == null ? "null" : bonfireInfoNodes.getLength()) +
                                        " <jFedBonfire:bonfire_node_info> elements");

                            if (!usedBonfireInfoNode) {
                                String publicNetId = fixedPublicNetworkIdForLocation(computeInfo.locationName);
                                String wanNetId = fixedWanNetworkIdForLocation(computeInfo.locationName);
//                                if (publicNetId != null) computeInfo.networkIds.add(publicNetId); //TODO this should not be default
                                if (wanNetId != null) computeInfo.networkIds.add(wanNetId);
                                computeInfo.storageIds.add(fixedStorageIdForLocation(computeInfo.locationName));
                                instanceType = "lite";
                                LOG.warn("No bonfire node info used. Reverting to defaults!");
                            }

                            RestApi.RestReply<String> reply = occi.postExperimentComputeWizard(con, exInf.experimentId,
                                    computeInfo.computeName,
                                    groups,
                                    computeInfo.locationName,
                                    instanceType,
                                    computeInfo.networkIds,
                                    computeInfo.storageIds);

                            if (!reply.getGeniResponseCode().isSuccess())
                                throw new JFedException("Could not post experiment compute", null, reply.getGeniResponseCode());

                            ComputeInfo computeInfoRecv = parseCompute(reply.getValue());

                            fakeManifestRspecBuilder.append(
                                    " <node client_id=\"" + computeInfo.computeName + "\" " +
                                            "       component_manager_id=\"" + componentManagerId + "\" " +
                                            "       exclusive=\"false\" " +
                                            "       component_id=\"" + componentId + "\"" +
                                            "       sliver_id=\"" + exInf.sliverUrn.toString() + "\"> " +
                                            "   <sliver_type name=\"" + instanceType + "\"/> "
                            );

                            String publicNetId = fixedPublicNetworkIdForLocation(computeInfo.locationName);
                            String wanNetId = fixedWanNetworkIdForLocation(computeInfo.locationName);
                            if (computeInfoRecv.ipsByNetworkId.containsKey(publicNetId) || computeInfoRecv.ipsByNetworkId.containsKey(wanNetId)) {
                                String nodename;
                                if (computeInfoRecv.ipsByNetworkId.containsKey(publicNetId))
                                    nodename = computeInfoRecv.ipsByNetworkId.get(publicNetId);
                                else
                                    nodename = computeInfoRecv.ipsByNetworkId.get(wanNetId);

                                LOG.debug("Using IP in bonfire compute info for SSH: " + nodename);

                                assert nodename != null;
                                String port = "22";
                                String username = "root";
                                fakeManifestRspecBuilder.append(
                                        "   <services>" +
                                                "       <login authentication=\"ssh-keys\" " +
                                                "hostname=\"" + nodename + "\" " +
                                                "port=\"" + port + "\" " +
                                                "username=\"" + username + "\"/>" +
                                                "   </services>"
                                );
                            } else {
                                LOG.debug("No IP in bonfire compute info");
                            }
                            fakeManifestRspecBuilder.append("</node> ");
                        }
                    }
                }
            }
        } catch (ParserConfigurationException e) {
            throw new RuntimeException("ParserConfigurationException: " + e.getMessage(), e);
        } catch (SAXException e) {
            throw new RuntimeException("SAXException: " + e.getMessage(), e);
        } catch (IOException e) {
            throw new RuntimeException("IOException: " + e.getMessage(), e);
        } catch (GeniUrn.GeniUrnParseException e) {
            throw new RuntimeException("GeniUrnParseException: " + e.getMessage(), e);
        }

        fakeManifestRspecBuilder.append("</rspec>");

        exInf.manifestRspec = fakeManifestRspecBuilder.toString();


        //HACK to send manifest to sfaModel if any
        Map<String, Object> methodParameters = new HashMap<String, Object>();
        methodParameters.put("sliceUrn", exInf.sliceUrn);
        methodParameters.put("sliverUrn", exInf.sliverUrn);
        methodParameters.put("fakeManifest", exInf.manifestRspec);
        ApiCallDetails fakeManifestCall = new ApiCallDetails(
                amAuthority, null/*connectionConfig*/, null/*serverUrl*/, "fake",
                OCCI.getApiName(), "fakeManifest", "fakeManifest",
                null, null, null, null, null, null, null, null, null,
                methodParameters,
                null, null, null
        );
        occi.getLogger().fireResult(fakeManifestCall);
        //end HACK


        return exInf.manifestRspec;
    }

    @Override
    public boolean hasSeperateAllocateAndProvision() {
        return false;
    }

    @Override
    public AllocateReply allocate(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, Date expirationDate) throws JFedException, VlanUnavailableException, AnyVlanUnsupportedException {
        throw new RuntimeException("Unsupported operation");
    }

    @Override
    public String provisionAndStart(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<UserSpec> userSpecs, Date expirationDate) throws JFedException, VlanUnavailableException, AnyVlanUnsupportedException {
        throw new RuntimeException("Unsupported operation");
    }

    @Override
    public void deleteSliver(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        HttpConnection con;
        try {
            con = getConnection();
        } catch (ExtraInfoCallback.NoBonfireAccountException e) {
            return;
        }

        //delete bonfire experiment related to this sliver
        ExperimentInfo exInf = sliceUrnToExperimentInfo.get(sliceUrn);

        LOG.debug("OcciToAggregateManagerWrapper.deleteSliver(" + sliceUrn + ") -> exInf=" + exInf);

        if (exInf != null && exInf.experimentId != null) {
            LOG.debug("OcciToAggregateManagerWrapper.deleteSliver(" + sliceUrn + ") is deleting -> exInf.experimentId=" + exInf.experimentId);
            RestApi.RestReply<String> reply = occi.deleteExperiment(con, exInf.experimentId);

            if (reply.getGeniResponseCode().isSuccess())
                sliceUrnToExperimentInfo.remove(sliceUrn);
        } else {
            LOG.debug("Could not find experiment to delete! ");
        }
    }

    @Override
    public StatusDetails status(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        HttpConnection con;
        try {
            con = getConnection();
        } catch (ExtraInfoCallback.NoBonfireAccountException e) {
            return null;
        }

        //check status of experiment computes
        ExperimentInfo exInf = sliceUrnToExperimentInfo.get(sliceUrn);

        if (exInf != null && exInf.experimentId != null) {
            RestApi.RestReply<String> replyComputes = occi.getExperimentComputes(con, exInf.experimentId);

            if (replyComputes.getGeniResponseCode().isSuccess()) {
                List<ComputeInfo> computeInfos = parseExperimentComputes(exInf.experimentId, replyComputes.getValue());

                LOG.debug("OcciToAggregateManagerWrapper.status checking " + computeInfos.size() + " computes");

                boolean allReady = true;
                for (ComputeInfo computeInfo : computeInfos) {
                    RestApi.RestReply<String> replyLocCompute =
                            occi.getLocationCompute(con, computeInfo.locationName, computeInfo.computeId);

                    if (!replyLocCompute.getGeniResponseCode().isSuccess()) {
                        LOG.debug("OcciToAggregateManagerWrapper.status   failed to get status for compute " + computeInfo + ". status unknown. " + replyLocCompute.getGeniResponseCode());
                        return new StatusDetails(StatusDetails.SliverStatus.UNKNOWN); //fail or unallocated but we don't know...
                    } else {
                        OcciToAggregateManagerWrapper.ComputeInfo infoRecv =
                                OcciToAggregateManagerWrapper.parseCompute(replyLocCompute.getValue());
                        LOG.debug("OcciToAggregateManagerWrapper.status got status " + infoRecv.state + " (ready=" + infoRecv.ready() + ") for compute " + computeInfo);
                        if (!infoRecv.ready())
                            allReady = false;
                    }
                }
                LOG.debug("OcciToAggregateManagerWrapper.status   allready=" + allReady);

                if (allReady)
                    return new StatusDetails(StatusDetails.SliverStatus.READY);
                else
                    return new StatusDetails(StatusDetails.SliverStatus.CHANGING);
            }
        }

        return new StatusDetails(StatusDetails.SliverStatus.UNKNOWN); //fail or unallocated but we don't know...
    }

    @Override
    public String describe(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        //generate RSpec based on experiment  This is actually done on createSliver, and stored version is returned here.
        ExperimentInfo exInf = sliceUrnToExperimentInfo.get(sliceUrn);

        if (exInf != null) {
            //totally fake
            return exInf.manifestRspec;
        }

        return null;
    }

    @Override
    public void renewSlivers(GeniUrn sliceUrn, List<GeniUrn> sliverUrns, List<AnyCredential> sliceCredentials, Date newExpire) throws JFedException {
        //modify bonfire experiment wall time
        ExperimentInfo exInf = sliceUrnToExperimentInfo.get(sliceUrn);

        throw new RuntimeException("Not yet implemented");
    }

    @Override
    public boolean isExirationdateSetByCreateSliver() {
        return true;
    }

    @Override
    public boolean isShareLanSupported() {
        return false;
    }

    @Override
    public void shareLan(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String lanClientId, String newSharedLanName) throws JFedException {
        throw new RuntimeException("not supported");
    }

    @Override
    public void unshareLan(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String lanClientId, String sharedLanName) throws JFedException {
        throw new RuntimeException("not supported");
    }

    @Override
    public void restart(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        throw new RuntimeException("not supported");
    }

    @Override
    public void reloadOS(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        throw new RuntimeException("not supported");
    }

    @Override
    public URL openConsole(GeniUrn sliverUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        throw new RuntimeException("not supported");
    }

    @Override
    public boolean isRestartSupported() {
        return false;
    }

    @Override
    public boolean isReloadOSSupported() {
        return false;
    }

    @Override
    public boolean isOpenConsoleSupported() {
        return false;
    }

    @Override
    public boolean isEditSshKeysSupported() {
        return false;
    }

    @Override
    public void editSshKeys(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, Collection<UserSpec> userSpecs) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }

    @Override
    public boolean areUserDiskImageFunctionsSupported() {
        return false;
    }

    @Override
    public ImageInfo createImage(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, GeniUrn componentUrn, String imageName, boolean global, boolean updatePrepare) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }

    public static class ExperimentInfo {
        public GeniUrn sliceUrn;
        public GeniUrn sliverUrn;
        public String experimentName;
        public String experimentId;
        public List<ComputeInfo> computes;
        public String manifestRspec;

        public ExperimentInfo(GeniUrn sliceUrn, GeniUrn sliverUrn, String experimentName) {
            this.sliceUrn = sliceUrn;
            this.sliverUrn = sliverUrn;
            this.experimentName = experimentName;
        }

        @Override
        public String toString() {
            return "ExperimentInfo{" +
                    "sliceUrn=" + sliceUrn +
                    ", experimentId='" + experimentId + '\'' +
                    '}';
        }
    }

    public static class ComputeInfo {
        public final List<String> storageIds = new ArrayList<String>();
        public final List<String> networkIds = new ArrayList<String>();
        public final Map<String, String> ipsByNetworkId = new HashMap<String, String>();
        public String experimentId;
        public String computeName;
        public String computeId;
        public String locationName;
        public String state;

        public boolean ready() {
            if (state == null) return false;
            return state.equalsIgnoreCase("ACTIVE") || state.equalsIgnoreCase("RUNNING") || state.equalsIgnoreCase("READY");
        }

        @Override
        public String toString() {
            return "ComputeInfo{" +
                    "computeId='" + computeId + '\'' +
                    ", locationName='" + locationName + '\'' +
                    ", experimentId='" + experimentId + '\'' +
                    '}';
        }
    }

    public static abstract class HrefAndNameInfo {
        protected final String href;
        protected final String name;

        protected HrefAndNameInfo(String href, String name) {
            this.href = href;
            this.name = name;
        }

        public String getHref() {
            return href;
        }

        public String getName() {
            return name;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            HrefAndNameInfo that = (HrefAndNameInfo) o;

            if (!href.equals(that.href)) return false;
            if (!name.equals(that.name)) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = href.hashCode();
            result = 31 * result + name.hashCode();
            return result;
        }
    }

    public static class LocationInfo extends HrefAndNameInfo {
        private final String url;

        public LocationInfo(String href, String name, String url) {
            super(href, name);
            this.url = url;
        }

        public String getUrl() {
            return url;
        }

        @Override
        public String toString() {
            return "LocationInfo{href='" + href + '\'' + ", name='" + name + '\'' + ", url='" + url + '\'' + '}';
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;
            if (!super.equals(o)) return false;

            LocationInfo that = (LocationInfo) o;

            if (!url.equals(that.url)) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = super.hashCode();
            result = 31 * result + url.hashCode();
            return result;
        }
    }

    public static class StorageInfo extends HrefAndNameInfo {
        protected StorageInfo(String href, String name) {
            super(href, name);
        }
    }

    public static class NetworkInfo extends HrefAndNameInfo {
        protected NetworkInfo(String href, String name) {
            super(href, name);
        }
    }
}
