package be.iminds.ilabt.jfed.lowlevel.stitching;

import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.BasicStringRspec;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.RFC3339Util;
import be.iminds.ilabt.jfed.util.XmlUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.annotation.Nullable;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

/**
 * StitchingData stores StitchingHopData and StitchingCallData
 * <p/>
 * It also knows the order in which AM's need to be called.
 * <p/>
 * It derives all data from the info returned from the SCS call:
 * - stitching extension of returned rspec
 * - workflow data
 */
public class StitchingData {
    private final org.slf4j.Logger log;

    public final static String NAMESPACE_STITCHING = "http://hpn.east.isi.edu/rspec/ext/stitch/0.1/";
    public final static String NAMESPACE_RSPEC = "http://www.geni.net/resources/rspec/3";

    private final List<StitchingHopData> stitchingHopDataList = new ArrayList<StitchingHopData>();
    private final Map<String, StitchingCallData> stitchingCallDataByAuthUrn = new HashMap<String, StitchingCallData>();

    private final List<StitchingCallData> orderedStitchingCallDataList = new ArrayList<StitchingCallData>();
    private final List<SfaAuthority> orderedSfaAuths = new ArrayList<SfaAuthority>();

    private final Set<SfaAuthority> fakeSfaAuths = new HashSet<SfaAuthority>();

    private final String scsReplyRspec;

    public StitchingData(String scsReplyRspec, Hashtable workflow, AuthorityListModel authList, AuthorityFinder authFinder, Logger log) {
        if (log == null)
            this.log = LoggerFactory.getLogger(StitchingData.class);
        else
            this.log = log;
        this.scsReplyRspec = scsReplyRspec;

        Element stitchingElement = extractStitchingElementFromRspec(scsReplyRspec, log);
        assert stitchingElement != null : "No stitching element in rspec";

        //then go over each hop in the workflow
        Hashtable<String, Hashtable> workflowData = getAsHtofStoHt(workflow);
        Map<String, String> aggUrnsToAggUrl =
                new HashMap<String, String>(); //keep track of all unique aggregates
        for (String linkName : workflowData.keySet()) {
            Vector<Hashtable> linkDeps = getValueOfTypeVofHt(workflowData.get(linkName), "dependencies");
            assert linkDeps != null;

            List<Hashtable> hops = new LinkedList<Hashtable>();
            hops.addAll(linkDeps);
            while (!hops.isEmpty()) {
                Hashtable hopHt = hops.remove(0);

                if (hopHt.containsKey("dependencies")) {
                    Vector<Hashtable> deps = getValueOfTypeVofHt(hopHt, "dependencies");
                    hops.addAll(deps);
                }

                //this is incorrect, if we do this here, stitching data gets removed from the rspec!
//                String aggregateUrn = (String) hopHt.get("aggregate_urn");
//                SfaAuthority hopAggAuth = authList.getFromAnyUrn(aggregateUrn);
//                if (aggregateUrn != null && hopAggAuth != null && hopAggAuth.getType().equals("fake")) {
//                    //ignore this hop because it is a fake hop used for adding networks
//                    continue;
//                }

                String aggregateUrn = (String) hopHt.get("aggregate_urn");


//                SfaAuthority hopAggAuth = authList.getByUrnExact(aggregateUrn); //same but with finder:
                SfaAuthority hopAggAuth = authFinder.findByUrn(aggregateUrn, AuthorityFinder.Purpose.STITCHING);
                if (hopAggAuth == null) {
                    String aggregateUrlString = (String) hopHt.get("aggregate_url");

                    //dirty exception to fix SCS error
                    if (aggregateUrlString.equals("https://uvanl-hn.exogeni.net:11443/orca/xmlrpc")) {
                        aggregateUrlString = "https://uva-nl-hn.exogeni.net:11443/orca/xmlrpc";
                    }

                    //fallback to using URL
                    log.debug("Did not find authority \"" + aggregateUrn + "\" by using URN. Falling back to searching for URL \"" + aggregateUrlString + "\"");
                    if (aggregateUrlString != null)
                        try {
                            URL aggregateUrl = new URL(aggregateUrlString);
                            hopAggAuth = authList.getByUrl(aggregateUrl);
                            if (hopAggAuth != null)
                                log.debug("Found auth by URL. Real URN is: \""+hopAggAuth.getUrnString()+"\"");
                        } catch (MalformedURLException e) {
                            log.warn("Malformed URL in workflow aggregate_url: \"" + aggregateUrlString + "\"", e);
                        }
                }

                if (hopAggAuth == null) {
                    throw new RuntimeException("Could not find the authority \""+aggregateUrn+"\" which is described in the stitching workflow. Giving up.");
                }

                if (!hopAggAuth.getUrnString().equals(aggregateUrn)) {
                    log.debug("Hop with auth " + aggregateUrn + " will be seen as hop with auth " + hopAggAuth.getUrnString());
                    //force same as auth. This is important in the exogeni case. WE might need to use exoSM or a specific rack. The SCS decides!
                    aggregateUrn = hopAggAuth.getUrnString();
                    hopHt.put("aggregate_urn", aggregateUrn);
                } else
                    log.debug("Hop with auth "+aggregateUrn+"");

                if (aggregateUrn != null && hopAggAuth != null && hopAggAuth.getType() != null && hopAggAuth.getType().equals("fake")) {
                    fakeSfaAuths.add(hopAggAuth);
                }

                StitchingHopData stitchingHopData = new StitchingHopData(linkName, hopHt, stitchingElement, log);
                if (stitchingHopDataList.contains(stitchingHopData)) {
                    int i = stitchingHopDataList.indexOf(stitchingHopData);
                    assert i >= 0;
                    assert i < stitchingHopDataList.size();
                    stitchingHopData = stitchingHopDataList.get(i); //replace created stitchingHopData with existing
                    stitchingHopData.processExtraWorkflowData(linkName, hopHt);
                } else {
                    stitchingHopDataList.add(stitchingHopData);
                }

                assert stitchingHopData.getAggregateUrn().equals(hopAggAuth.getUrnString());
                assert stitchingHopData.getAggregateUrn().equals(aggregateUrn);

                if (!aggUrnsToAggUrl.containsKey(stitchingHopData.getAggregateUrn()))
                    aggUrnsToAggUrl.put(stitchingHopData.getAggregateUrn(), stitchingHopData.getAggregateUrl());
            }
        }
        //link dependency graph of stitchingHopDataList
        for (StitchingHopData stitchingHopData : stitchingHopDataList) {
            stitchingHopData.linkDeps(stitchingHopDataList);
        }

        log.debug("stitchingHopDataList=" + stitchingHopDataList);

        //we now have a list with all StitchingHopData
        //we will create all StitchingCallData
        Set<String> stitchedAuthorityUrns = new HashSet<String>(); //also keep track of all authorities
        for (Map.Entry<String, String> e : aggUrnsToAggUrl.entrySet()) {
            String urn = e.getKey();
            String url = e.getValue();

            StitchingCallData stitchingCallData = new StitchingCallData(
                    urn, url,
                    authList, log);
            stitchingCallDataByAuthUrn.put(urn, stitchingCallData);
            stitchedAuthorityUrns.add(urn);
        }
        for (StitchingHopData stitchingHopData : stitchingHopDataList) {
            StitchingCallData stitchingCallData = stitchingCallDataByAuthUrn.get(stitchingHopData.getAggregateUrn());
            assert stitchingCallData != null;
            stitchingCallData.addHopData(stitchingHopData);
        }
        for (StitchingCallData stitchingCallData : stitchingCallDataByAuthUrn.values()) {
            stitchingCallData.linkDeps(stitchingCallDataByAuthUrn.values());
        }

        log.debug("stitchedAuthorityUrns=" + stitchedAuthorityUrns);
        log.debug("stitchingCallDataByAuthUrn=" + stitchingCallDataByAuthUrn);

        //map how exogeni is used in the workflow
        boolean stitchedExoGeniToplevelAuth = false;
        Set<GeniUrn> stitchedExoGeniSubAuths = new HashSet<GeniUrn>();
        for (StitchingHopData stitchingHopData : stitchingHopDataList) {
            GeniUrn hopAuthUrn = GeniUrn.parse(stitchingHopData.getAggregateUrn());
            if (hopAuthUrn != null && hopAuthUrn.getEncodedTopLevelAuthority_withoutSubAuth().equals("exogeni.net")) {
                if (hopAuthUrn.hasSubAuthority())
                    stitchedExoGeniSubAuths.add(hopAuthUrn);
                else
                    stitchedExoGeniToplevelAuth = true;
            }
        }

        if (!stitchedExoGeniSubAuths.isEmpty() && stitchedExoGeniToplevelAuth)
            throw new RuntimeException("Exogeni stitching with a too complex scenario: both sub authorities and top level authority need to be involved. jFed cannot handle this.");

        //now create and add the StitchingCallData for AM's that are NOT involved in stitching at all, but are in the RSpec.
        List<String> allAuthorityUrns = new BasicStringRspec(scsReplyRspec).getAllComponentManagerUrns();
        Set<String> nonStitchedAuthorityUrns = new HashSet<String>(allAuthorityUrns);
        for (String urn : allAuthorityUrns) {
            GeniUrn u = GeniUrn.parse(urn);
            assert u != null;
            if (u != null) {
                if (u.getEncodedTopLevelAuthority_withoutSubAuth().equals("exogeni.net")) {

                    //ignore componentmanagers like urn:publicid:IDN+exogeni.net:bbnNet+authority+am
                    if (u.hasSubAuthority() && u.getEncodedSubAuthName().endsWith("Net")) {
                        nonStitchedAuthorityUrns.remove(urn);
                        continue;
                    }

                    if (!u.hasSubAuthority()) {
                        //exogeni.net top level itself
                        if (stitchedExoGeniToplevelAuth) {
                            //nothing to do
                        } else {
                            if (!stitchedExoGeniSubAuths.isEmpty())
                                throw new RuntimeException("Exogeni stitching with a too complex scenario: both sub authorities and top level authority need to be involved. jFed cannot handle this.");

                            stitchedExoGeniToplevelAuth = true;
                            nonStitchedAuthorityUrns.add(u.toString());
                        }
                    } else {
                        //not that this is a tricky situation: we can't use the user exogeni preferences, as we can't mix sub auth and top level auth.
                        if (!stitchedExoGeniToplevelAuth) {
                            if (!stitchedExoGeniSubAuths.contains(u)) {
                                //involve sub auth
                                stitchedExoGeniSubAuths.add(u);
                                nonStitchedAuthorityUrns.add(u.toString());
                            }
                        } else {
                            //nothing to do: we DON't want to involve any sub auth, the top auth will handle it.
                        }
                    }
                }
            }

        }
        nonStitchedAuthorityUrns.removeAll(stitchedAuthorityUrns);
        log.debug("nonStitchedAuthorityUrns=" + nonStitchedAuthorityUrns);
        for (String nonStitchedAuthorityUrn : nonStitchedAuthorityUrns) {
            StitchingCallData stitchingCallData = new StitchingCallData(
                    nonStitchedAuthorityUrn,
                    null /*url not needed if urn is of known authority (and it should be or we can't use this RSpec even without stitching)*/,
                    authList,
                    log);
            stitchingCallDataByAuthUrn.put(nonStitchedAuthorityUrn, stitchingCallData);
        }


        //now find the ordering: copy depgraph and keep removing nodes without deps until nothing is left. The order of removing is the order searched.
        Map<StitchingCallData, StitchingCallDataDepGrahpHelper> depGraphCopy =
                new HashMap<StitchingCallData, StitchingCallDataDepGrahpHelper>();
        for (StitchingCallData stitchingCallData : stitchingCallDataByAuthUrn.values())
            depGraphCopy.put(stitchingCallData, new StitchingCallDataDepGrahpHelper(stitchingCallData));
        if (!stitchingCallDataByAuthUrn.isEmpty())
            while (orderedStitchingCallDataList.size() != stitchingCallDataByAuthUrn.size()) {
                int prevOrderedStitchingCallDataListSize = orderedStitchingCallDataList.size();

                for (StitchingCallDataDepGrahpHelper h : depGraphCopy.values()) {
                    if (h.dependencies.isEmpty() && !orderedStitchingCallDataList.contains(h.s)) {
                        for (StitchingCallData dep : h.dependingOnThis) {
                            StitchingCallDataDepGrahpHelper hDep = depGraphCopy.get(dep);
                            hDep.dependencies.remove(h.s);
                        }
                        h.dependingOnThis.clear();
                        orderedStitchingCallDataList.add(h.s);
                    }
                }

                if (prevOrderedStitchingCallDataListSize == orderedStitchingCallDataList.size())
                    throw new RuntimeException("Loop in depgraph (size stays "+prevOrderedStitchingCallDataListSize+" targetsize="+stitchingCallDataByAuthUrn.size()+"): " +
                            "unlooped: "+orderedStitchingCallDataList+" all: "+depGraphCopy.values());
            }

        //create orderedSfaAuths
        boolean sawExoGeni = false;
        List<GeniUrn> orderedSfaAuthUrns = new ArrayList<GeniUrn>();
        for (StitchingCallData dep : orderedStitchingCallDataList) {
            assert !orderedSfaAuths.contains(dep.getAuth()) : "duplicate auth: "+orderedSfaAuths+" + "+dep.getAuth();
            boolean isExoGeni = dep.getAuth().getUrn().getEncodedTopLevelAuthority_withoutSubAuth().equals("exogeni.net");
            if (isExoGeni) {
                assert !sawExoGeni : "more then one exogeni auth: "+orderedSfaAuths+" + "+dep.getAuth();
                sawExoGeni = true;
            }

            orderedSfaAuths.add(dep.getAuth());
            orderedSfaAuthUrns.add(dep.getAuth().getUrn());
        }
        log.debug("orderedSfaAuthUrns=" + orderedSfaAuthUrns);
    }


    public static Element extractStitchingElementFromRspec(String rspec) {
        return extractStitchingElementFromRspec(rspec, null);
    }
    public static Element extractStitchingElementFromRspec(String rspec, @Nullable Logger log) {
        if (log == null)
            log = LoggerFactory.getLogger(StitchingData.class);
        try {
            //first parse rspec
            Document doc = XmlUtil.parseXmlFromString(rspec, true/*namespaceaware*/);

            doc.getDocumentElement().normalize();

            NodeList rspecList;
            NodeList rspecListNS = doc.getElementsByTagNameNS(NAMESPACE_RSPEC, "rspec");
            if (rspecListNS.getLength() == 0) {
                NodeList rspecListNoNS = doc.getElementsByTagName("rspec");
                rspecList = rspecListNoNS;
            } else
                rspecList = rspecListNS;

            assert rspecList.getLength() == 1 :
                    "not a single rspec element, but " + rspecList.getLength() + " in " + rspec;
            Node rspecNode = rspecList.item(0);
            assert rspecNode != null;
            assert rspecNode.getNodeType() == Node.ELEMENT_NODE;
            Element rspecElement = (Element) rspecNode;

            NodeList stitchingList;
            NodeList stitchingListNS = rspecElement.getElementsByTagNameNS(NAMESPACE_STITCHING, "stitching");
            if (stitchingListNS.getLength() == 0) {
                //fallback to looking for stitching element without namespace.
                NodeList stitchingListNoNS = rspecElement.getElementsByTagName("stitching");
                if (stitchingListNoNS.getLength() == 0)
                    return null;
                stitchingList = stitchingListNoNS;
            } else
                stitchingList = stitchingListNS;

            assert stitchingList.getLength() == 1 :
                    "not a single stitching element, but " + stitchingList.getLength() + " in " + rspec;
            Node stitchingNode = stitchingList.item(0);
            assert stitchingNode != null;
            assert stitchingNode.getNodeType() == Node.ELEMENT_NODE;
            Element stitchingElement = (Element) stitchingNode;

            return stitchingElement;
        } catch (SAXException e) {
            log.error("Error parsing rspec", e);
        } catch (IOException e) {
            log.error("Error parsing rspec", e);
        }
        return null;
    }

    /**
     * Get the current Rspec, including updated suggestVlan and availableVLAN data.
     * <p/>
     * This is based on the original request RSpec returned by the SCS, combined with updates for
     * availableVLAN and suggestedVLAN from "vlan unavailable" errors and CreateSliver manifests
     *
     * This also adds or removes the "expires" attribute of the RSpec tag, as needed.
     */
    public String getCurrentRequestRspec(Collection<GeniUrn> aggUrnsNotSupportingAny, boolean useExpires, @Nullable Date expireDate) {
        log.trace("getCurrentRequestRspec");
        //rewrite rspec, so that vlanInfo is used instead of what is in the rspec (requires DOM + transformation)

        try {
            DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
            dbFactory.setNamespaceAware(true);
            DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
            Document doc = dBuilder.parse(new InputSource(new StringReader(scsReplyRspec)));
            assert doc != null;

            doc.getDocumentElement().normalize();

            Element root = doc.getDocumentElement();

            //rspec - stitching - path - hop (MULTIPLE) - link - switchingCapabilityDescriptor - switchingCapabilitySpecificInfo - switchingCapabilitySpecificInfo_L2sc -
            //          vlanRangeAvailability
            //          suggestedVLANRange

            NodeList rspecList = doc.getElementsByTagNameNS(NAMESPACE_RSPEC, "rspec");
            assert rspecList.getLength() == 1 :
                    "not a single rspec element, but " + rspecList.getLength() + " in " + scsReplyRspec;
            Node rspecNode = rspecList.item(0);
            assert rspecNode != null;
            assert rspecNode.getNodeType() == Node.ELEMENT_NODE;
            Element rspecElement = (Element) rspecNode;

            if (useExpires) {
                if (expireDate != null)
                    rspecElement.setAttribute("expires", RFC3339Util.dateToRFC3339String(expireDate, true/*force Z*/));
            } else {
                rspecElement.removeAttribute("expires");
            }

            NodeList stitchingListNoNS = rspecElement.getElementsByTagNameNS(NAMESPACE_STITCHING, "stitching");
            assert stitchingListNoNS.getLength() == 1 :
                    "not a single stitching element, but " + stitchingListNoNS.getLength() + " in " + scsReplyRspec;
            Node stitchingNode = stitchingListNoNS.item(0);
            assert stitchingNode != null;
            assert stitchingNode.getNodeType() == Node.ELEMENT_NODE;
            Element stitchingElement = (Element) stitchingNode;

            List<Element> pathEls = XmlUtil.getDomElementChildElementsNS(stitchingElement, StitchingData.NAMESPACE_STITCHING, "path");
            for (Element pathEl : pathEls) {
                String linkName = pathEl.getAttribute("id");
                assert linkName != null : "<stitching> has <path> element without id attribute: " + pathEl;

                List<Element> origHopElements = XmlUtil.getDomElementChildElementsNS(pathEl, StitchingData.NAMESPACE_STITCHING, "hop");

                log.trace("getCurrentRequestRspec() rewriting path with link id=" + linkName + " orig el count: " + origHopElements.size());

                //remove original hop elements
                for (Element hopElement : origHopElements) {
                    hopElement.getParentNode().removeChild(hopElement);
                }

                //keep only hops for this path
                List<StitchingHopData> filteredHopData = new ArrayList<StitchingHopData>(stitchingHopDataList);
                for (ListIterator<StitchingHopData> it = filteredHopData.listIterator(); it.hasNext(); ) {
                    StitchingHopData hop = it.next();

                    if (!hop.getLinkName().equals(linkName))
                        it.remove();
                }

                //sort according to hop id (seems to be needed!)
                List<StitchingHopData> sortedHopData = new ArrayList<StitchingHopData>(filteredHopData);
                Collections.sort(sortedHopData, new Comparator<StitchingHopData>() {
                    @Override
                    public int compare(StitchingHopData o1, StitchingHopData o2) {
                        return o1.getRspecStitchingHopId().compareTo(o2.getRspecStitchingHopId());
                    }
                });

                Map<StitchingHopData, String> hopDataToFixedVlans = new HashMap<StitchingHopData, String>();
                for (StitchingHopData hop : sortedHopData) {
                    assert hop != null;
                    assert doc != null;
                    for (StitchingHopData depHop : hop.getDependencies()) {
                        if (depHop.getStitchingExtensionSource() == StitchingHopData.StitchingExtensionSource.MANIFEST) {
                            //if dependency is fixed, set suggested and available to only that!
                            Integer fixVlan = depHop.getSuggestedVlan();
                            if (fixVlan != null)
                                hopDataToFixedVlans.put(hop, fixVlan+"");
                        }
                    }
                }
                log.debug("hopDataToFixedVlans=" + hopDataToFixedVlans);

                //add updated hop elements
                for (StitchingHopData hop : sortedHopData) {
                    assert hop != null;
                    assert doc != null;

                    boolean useAny = true;

                    //not always so for all links:
//                    //if hop has deps, the dep has already chosen a vlan, and we don't get to choose
//                    if (! hop.getDependencies().isEmpty())
//                        useAny = false;

//                    for (StitchingHopData depHop : hop.getDependencies()) {
                    //why?
//                        if (depHop.getStitchingExtensionSource() == StitchingHopData.StitchingExtensionSource.MANIFEST)
//                            useAny = false;
                    //why?
//                        if (aggUrnsNotSupportingAny.contains(depHop.getAggregateUrn()+""))
//                            useAny = false;
//                    }

                    //why?
//                    if (hop.getStitchingExtensionSource() == StitchingHopData.StitchingExtensionSource.MANIFEST)
//                        useAny = false;

                    GeniUrn hopAggregateU = GeniUrn.parse(hop.getAggregateUrn());
                    if (hopAggregateU == null)
                        useAny = false; //safety fallback
                    else {
                        String hopAggregateTLA = hopAggregateU.getEncodedTopLevelAuthority_withoutSubAuth();
                        for (GeniUrn aggUrn : aggUrnsNotSupportingAny) {
                            if (aggUrn.getEncodedTopLevelAuthority_withoutSubAuth().equals(hopAggregateTLA))
                                useAny = false;
                        }
                    }
                    log.debug("useAny=" + useAny + " for " + hop.getAggregateUrn());

                    String fixedVlan = hopDataToFixedVlans.get(hop);
                    Element rspecStitchingHopElement;
                    if (fixedVlan != null) {
                        log.debug("using fixed vlan (" + fixedVlan + ") (due to dep) for " + hop.getAggregateUrn());
                        rspecStitchingHopElement = hop.getRspecStitchingHopElement(fixedVlan, fixedVlan);
                    }
                    else
                        rspecStitchingHopElement = hop.getRspecStitchingHopElement(useAny);
                    assert rspecStitchingHopElement != null;
                    Node importedNode = doc.importNode(rspecStitchingHopElement, true/*deep*/);
                    assert importedNode != null;

//                    logger.trace("getCurrentRequestRspec() adding hop   " + hop.getRspecStitchingHopId()+"  suggestedVlan="+hop.getSuggestedVlan()+"  nodeVlan={}",
//                            XmlUtil.elementToXmlString(hop.getRspecStitchingHopElement()));

//                    Element hopElement = (Element) importedNode;
//                    Element linkEl = XmlUtil.getDomElementFirstChildElement((hopElement, "link");
//                    Element scDescriptorEl = XmlUtil.getDomElementFirstChildElement((linkEl, "switchingCapabilityDescriptor");
//                    Element scSpecificInfoEl = XmlUtil.getDomElementFirstChildElement((scDescriptorEl, "switchingCapabilitySpecificInfo");
//                    Element scSpecificInfo_L2scEl = XmlUtil.getDomElementFirstChildElement((scSpecificInfoEl, "switchingCapabilitySpecificInfo_L2sc");
//                    Element vlanRangeAvailabilityEl = XmlUtil.getDomElementFirstChildElement((scSpecificInfo_L2scEl, "vlanRangeAvailability");
//                    Element suggestedVLANRangeEl = XmlUtil.getDomElementFirstChildElement((scSpecificInfo_L2scEl, "suggestedVLANRange");
//                    logger.trace("suggestedVLANRangeEl text="+suggestedVLANRangeEl.getTextContent());

//                    logger.trace("getCurrentRequestRspec() adding hop   " + hop.getRspecStitchingHopId()+"  suggestedVlan="+hop.getSuggestedVlan());
                    pathEl.appendChild(importedNode);
                }
            }

            try {
                TransformerFactory transFactory = TransformerFactory.newInstance();
                Transformer transformer = transFactory.newTransformer();
                StringWriter buffer = new StringWriter();
                transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
                transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
                transformer.setOutputProperty(OutputKeys.INDENT, "yes");
                transformer.transform(new DOMSource(rspecNode), new StreamResult(buffer));
//                logger.debug("generated RSPEC: {}", buffer.toString());
                return buffer.toString();
            } catch (TransformerConfigurationException e) {
                log.error("TransformerConfigurationException in getCurrentRequestRspec()", e);
            } catch (TransformerException e) {
                log.error("TransformerException in getCurrentRequestRspec()", e);
            }
        } catch (Exception e) {
            log.error("Exception in getCurrentRequestRspec()", e);
        }

        return null;
    }

    public List<StitchingCallData> getOrderedStitchingCallDataList() {
        return orderedStitchingCallDataList;
    }

    public StitchingCallData getStitchingCallData(String authUrn) {
        return stitchingCallDataByAuthUrn.get(authUrn);
    }

    public List<StitchingHopData> getStitchingHopDataList() {
        return stitchingHopDataList;
    }

    public List<SfaAuthority> getOrderedSfaAuths() {
        return orderedSfaAuths;
    }

    public Collection<SfaAuthority> getFakeSfaAuths() {
        return fakeSfaAuths;
    }


    @Override
    public String toString() {
        String orderedStitchingCallDataListString = "{";
        for (StitchingCallData stitchingCallData : orderedStitchingCallDataList)
            orderedStitchingCallDataListString += "\n    " + stitchingCallData;
        orderedStitchingCallDataListString += "}";

        return "StitchingData{" +
                "orderedStitchingCallDataList=" + orderedStitchingCallDataListString +
                '}';
    }

    public Hashtable<String, Hashtable> getAsHtofStoHt(Hashtable ht) {
        for (Object eo : ht.entrySet()) {
            Map.Entry e = (Map.Entry) eo;
            assert e != null;
            assert e.getKey() instanceof String :
                    "value in hashtable is not of type String, but of type " + e.getKey().getClass().getName();
            assert e.getValue() instanceof Hashtable :
                    "key in hashtable is not of type Hashtable, but of type " + e.getValue().getClass().getName();
        }

        return (Hashtable<String, Hashtable>) ht;
    }

    public Hashtable<String, Hashtable> getValueOfTypeHtofStoHt(Hashtable h, String key) {
        Object o = h.get(key);
        assert o != null : "Hashtable does not contain key \"" + key + "\". content: " + h;
        assert o instanceof Hashtable :
                "key in hashtable is not of type Hashtable, but of type " + o.getClass().getName();
        Hashtable ht = (Hashtable) o;
        return getAsHtofStoHt(ht);
    }

    public Vector<Hashtable> getValueOfTypeVofHt(Hashtable h, String key) {
        Object o = h.get(key);
        assert o != null : "Hashtable does not contain key \"" + key + "\". content: " + h;
        assert o instanceof Vector : "key in hashtable is not of type Vector, but of type " + o.getClass().getName();
        Vector v = (Vector) o;
        for (Object vo : v)
            assert vo instanceof Hashtable :
                    "key in hashtable is not of type Hashtable, but of type " + vo.getClass().getName();

        return (Vector<Hashtable>) v;
    }

    private static class StitchingCallDataDepGrahpHelper {
        List<StitchingCallData> dependencies = new ArrayList<StitchingCallData>();
        List<StitchingCallData> dependingOnThis = new ArrayList<StitchingCallData>();
        StitchingCallData s;

        public StitchingCallDataDepGrahpHelper(StitchingCallData s) {
            this.s = s;
            this.dependencies.addAll(s.getDependencies());
            this.dependingOnThis.addAll(s.getDependingOnThis());
        }

        @Override
        public String toString() {
            return "{"+s.getAuth()+" deps="+dependencies.size()+" depOnThis="+dependingOnThis.size()+"}";
        }
    }
}
