package be.iminds.ilabt.jfed.lowlevel.userloginmodel;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.PlanetlabSfaRegistryInterface;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.lowlevel.planetlab.PlanetlabCertificateFetcher;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.KeyUtil;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.*;
import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;
import java.util.ArrayList;
import java.util.List;

/**
 * This is a UserLoginModel that reads a key and certificate, and automatically extracts the authority and user URN.
 * <p/>
 * It also decodes and stores the private key using the password
 */
public class PlanetlabUserLoginModel implements UserLoginModel {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(PlanetlabUserLoginModel.class);
    private static final String PREFS_PLE_LOGIN_SOURCE = "pleLoginSource";
    private static final String PREFS_SSH_PRIVATE_KEY_FILE = "sshPrivateKeyFile";
    private static final String PREFS_SSH_PRIVATE_KEY_URL = "sshPrivateKeyUrl";
    private static final String PREFS_PLANETLAB_SFA_HRN = "planetlabSfaHrn";
    private static final String PREFS_PLANETLAB_AUTHORITY = "planetlabAuthority";
    private static final String PREFS_PLANETLAB_RETRIEVED_CERTIFICATE = "planetlabRetrievedCertificate";
    protected static File defaultSshPrivateKeyFile =
            new File(System.getProperty("user.home") + File.separator + ".ssh" + File.separator + "id_rsa");
    protected String planetlabSfaHrn;
    protected SfaAuthority authority;

    protected File sshPrivateKeyFile;
    protected URL sshPrivateKeyUrl;
    protected String privateKeyContent;
    protected LoginSource loginSource;


    protected List<X509Certificate> certificateChain;
    protected RSAPrivateKey privateKey; //unencrypted, so never saved directly!
    protected KeyPair keypair; //unencrypted, so never saved directly!
    protected GeniUrn userUrn;
    protected boolean passwordRequired;
    protected AuthorityListModel authorityListModel;
    protected UserLoginModelManager userLoginModelManager;
    private final Logger logger;
    private final JFedPreferences jFedPreferences;
    private final JFedConnectionProvider connectionProvider;

    public PlanetlabUserLoginModel(AuthorityListModel authorityListModel, UserLoginModelManager userLoginModelManager,
                                   Logger logger, JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences) {
        this.authorityListModel = authorityListModel;
        this.userLoginModelManager = userLoginModelManager;
        this.logger = logger;
        this.connectionProvider = connectionProvider;
        this.jFedPreferences = jFedPreferences;

        reset();
    }

    private static GeniUrn sfaHrnToUrn(String planetlabSfaHrn) {
        if (planetlabSfaHrn == null) return null;

        int lastDot = planetlabSfaHrn.lastIndexOf('.');
        assert lastDot != -1;
        String planetlabUsername = planetlabSfaHrn.substring(lastDot + 1);
        String planetlabAuthName = planetlabSfaHrn.substring(0, lastDot).replace('.', ':');

        return GeniUrn.createGeniUrnFromEncodedParts(planetlabAuthName, "user", planetlabUsername);
    }

    public void setAuthority(SfaAuthority auth) {
        this.authority = auth;
    }


    private void partialResetForLoad() {
        //partial reset
        this.loginSource = null;
        this.sshPrivateKeyFile = null;
        this.sshPrivateKeyUrl = null;
        this.passwordRequired = false;
        this.privateKeyContent = null;
        this.privateKey = null;
        this.keypair = null;
    }

    public void setSshPrivateKeyUrl(@Nonnull URL url) throws InvalidLoginException {
        partialResetForLoad();

        sshPrivateKeyUrl = url;
        loginSource = LoginSource.URL;

        try {
            URLConnection conn = url.openConnection();
            BufferedReader br = new BufferedReader(new InputStreamReader(conn.getInputStream()));

            StringBuilder sb = new StringBuilder();
            try {

                String line;
                while ((line = br.readLine()) != null) {
                    sb.append(line);
                    sb.append("\n");
                }
            } finally {
                br.close();
            }

            String newKeyCertContent = sb.toString();

            if (newKeyCertContent.isEmpty()) {
                throw new InvalidLoginException("Could not read planetlab login file from the provided URL. Got null");
            }

            this.privateKeyContent = newKeyCertContent;
            processLoginContent();

        } catch (IOException e) {
            LOG.error("Could not fetch planetlab login file", e);
            throw new InvalidLoginException("Could not fetch planetlab login file", e);
        }
    }


    public void setSshPrivateKeyFile(@Nonnull File newSshPrivateKeyFile) throws InvalidLoginException {
        partialResetForLoad();

        this.sshPrivateKeyFile = newSshPrivateKeyFile;
        loginSource = LoginSource.FILE;

        if (!newSshPrivateKeyFile.exists())
            throw new InvalidLoginException("SSH Private key file does not exist: \"" + newSshPrivateKeyFile.getPath() + "\"");
        this.sshPrivateKeyFile = newSshPrivateKeyFile;
        this.loginSource = LoginSource.FILE;

        try {
            this.privateKeyContent = IOUtils.fileToString(newSshPrivateKeyFile);
        } catch (IOException e) {
            throw new InvalidLoginException("Error reading \"" + newSshPrivateKeyFile.getPath() + "\"", e);
        }

        processLoginContent();

    }


    public void setSshPrivateKeyString(String newPrivateKeyContent) throws InvalidLoginException {
        partialResetForLoad();

        this.privateKeyContent = newPrivateKeyContent;
        this.loginSource = LoginSource.STRING;

        processLoginContent();
    }


    private void processLoginContent() throws InvalidLoginException {
        assert this.privateKeyContent != null;

        String separationChar = privateKeyContent.contains("\n") ? "\n"
                : " ";   //fallback for when enters are not present in login file

        boolean hasKey = KeyUtil.hasRsaPrivateKey(privateKeyContent);
        if (hasKey) {
            boolean encryptedKey = KeyUtil.hasEncryptedRsaPrivateKey(privateKeyContent);
            this.passwordRequired = encryptedKey;
            if (!encryptedKey)
                try {
                    this.keypair = KeyUtil.pemToRsaKeyPair(privateKeyContent, null);
                    this.privateKey = (RSAPrivateKey) keypair.getPrivate();
                } catch (KeyUtil.PEMDecodingException pemDecodingError) {
                    throw new InvalidLoginException("Error reading private key", pemDecodingError);
                }
        } else {
            this.privateKeyContent = null;
            throw new InvalidLoginException("Error no private key found in \"" + sshPrivateKeyFile.getPath() + "\"");
        }

        boolean containsSfaHrn = privateKeyContent.contains("#planetlab.sfa.hrn = ");
        if (containsSfaHrn) {
            int indexStart = privateKeyContent.indexOf("#planetlab.sfa.hrn = ") + "#planetlab.sfa.hrn = ".length();
            int indexStop = privateKeyContent.indexOf(separationChar, indexStart);
            String sfaHrn = privateKeyContent.substring(indexStart, indexStop).trim();
            LOG.trace("Found Sfa Hrn in private key file: '{}'", sfaHrn);
            setPlanetlabSfaHrn(sfaHrn);
        } else {
            if (privateKeyContent.startsWith("fed4fire."))
                containsSfaHrn = true;
            if (privateKeyContent.contains("http://sfa-fed4fire.pl.sophia.inria.fr:12345/"))
                containsSfaHrn = true;
            if (containsSfaHrn) {
                int indexStart = 0;
                int indexStop = privateKeyContent.indexOf(separationChar, indexStart);
                String sfaHrn = privateKeyContent.substring(indexStart, indexStop).trim();
                LOG.trace("Found fed4fire Sfa Hrn in private key file: '{}'", sfaHrn);
                setPlanetlabSfaHrn(sfaHrn);

                if (!correctAuthority()) {
                    SfaAuthority fed4fireAuth =
                            authorityListModel.getByUrnExact("urn:publicid:IDN+fed4fire+authority+sa");
                    if (fed4fireAuth == null)
                        throw new NullPointerException("fed4fire authority is not known: cannot continue.");
                    setAuthority(fed4fireAuth);
                }
            } else {
                LOG.trace("No Sfa Hrn in private key file");
            }
        }
    }

    public void setPlanetlabSfaHrn(String planetlabSfaHrn) {
        this.planetlabSfaHrn = planetlabSfaHrn;

        if (correctPlanetlabSfaUrn() && !correctAuthority()) {
            //see if we can find the authority automatically
            GeniUrn u = sfaHrnToUrn(planetlabSfaHrn);
            if (u.getEncodedTopLevelAuthority().startsWith("ple:") || u.getEncodedTopLevelAuthority().equals("ple")) {
                SfaAuthority authority2 =
                        authorityListModel.getByUrnExact("urn:publicid:IDN+" + u.getEncodedTopLevelAuthority() + "+authority+cm");
                if (authority2 != null) {
                    authority = authority2;
                    return;
                }
                SfaAuthority authority1 =
                        authorityListModel.getByUrnExact("urn:publicid:IDN+ple+authority+cm");
                if (authority1 != null) {
                    authority = authority1;
                }
            }
        }
    }

    public void reset() {
        this.planetlabSfaHrn = null;
        this.authority = null;
        this.sshPrivateKeyUrl = null;
        this.sshPrivateKeyFile = null;
        this.privateKeyContent = null;
        //derived
        this.certificateChain = null;

        this.passwordRequired = false;


        this.userUrn = null;
        this.privateKey = null;
        this.keypair = null;
    }

    public void defaults() {
        reset();
        //setSshPrivateKeyFile(defaultSshPrivateKeyFile);
    }

    public void save(UserLoginModelManager.UserLoginModelPreferences prefs) {
        assert (loginSource != null);

        prefs.put(PREFS_PLE_LOGIN_SOURCE, loginSource.name());

        if (sshPrivateKeyUrl != null) {
            prefs.put(PREFS_SSH_PRIVATE_KEY_URL, sshPrivateKeyUrl.toExternalForm());
        } else {
            prefs.remove(PREFS_SSH_PRIVATE_KEY_URL);
        }

        if (sshPrivateKeyFile != null) {
            prefs.put(PREFS_SSH_PRIVATE_KEY_FILE, sshPrivateKeyFile.toURI().toString());
        } else {
            prefs.remove(PREFS_SSH_PRIVATE_KEY_FILE);
        }
        if (planetlabSfaHrn != null) {
            prefs.put(PREFS_PLANETLAB_SFA_HRN, planetlabSfaHrn);
        } else {
            prefs.remove(PREFS_PLANETLAB_SFA_HRN);
        }
        if (authority != null) {
            prefs.put(PREFS_PLANETLAB_AUTHORITY, authority.getUrnString());
        } else {
            prefs.remove(PREFS_PLANETLAB_AUTHORITY);
        }
        if (certificateChain != null && !certificateChain.isEmpty()) {
            prefs.put(PREFS_PLANETLAB_RETRIEVED_CERTIFICATE, KeyUtil.x509certificateChainToPem(certificateChain));
        } else {
            prefs.remove(PREFS_PLANETLAB_RETRIEVED_CERTIFICATE);
        }
    }

    public void load(UserLoginModelManager.UserLoginModelPreferences prefs) throws InvalidLoginException {
        reset();

        planetlabSfaHrn = prefs.get(PREFS_PLANETLAB_SFA_HRN, null);
        if (planetlabSfaHrn != null) {
            if (!correctPlanetlabSfaUrn()) {
                LOG.warn("Stored planetlabSfaHrn not valid: {}", planetlabSfaHrn);
                prefs.remove(PREFS_PLANETLAB_SFA_HRN);
            }
            userUrn = sfaHrnToUrn(planetlabSfaHrn);
        }

        String planetlabAuthorityUrn = prefs.get(PREFS_PLANETLAB_AUTHORITY, null);
        authority =
                authorityListModel.getByUrnExact(planetlabAuthorityUrn);

        String planetlabRetreivedCertificatePem = prefs.get(PREFS_PLANETLAB_RETRIEVED_CERTIFICATE, null);
        if (planetlabRetreivedCertificatePem != null) {
            certificateChain = KeyUtil.pemToX509CertificateChain(planetlabRetreivedCertificatePem);
            if (certificateChain == null || certificateChain.isEmpty()) {
                LOG.warn("Stored certificate PEM is not valid: {}", planetlabRetreivedCertificatePem);
                prefs.remove(PREFS_PLANETLAB_RETRIEVED_CERTIFICATE);
            }
        }

        if (prefs.contains(PREFS_PLE_LOGIN_SOURCE)) {

            if (prefs.get(PREFS_PLE_LOGIN_SOURCE).equals(LoginSource.FILE.name())) {

                String uri = prefs.get(PREFS_SSH_PRIVATE_KEY_FILE, defaultSshPrivateKeyFile.toURI().toString());

                File file = null;
                try {
                    file = new File(new URI(uri));
                } catch (URISyntaxException e) {
                    LOG.warn("Stored file URI is not a valid URI. It will be deleted from the settings", e);
                    prefs.remove(PREFS_SSH_PRIVATE_KEY_FILE);

                }

                if (file != null) setSshPrivateKeyFile(file);
            } else if (prefs.get(PREFS_PLE_LOGIN_SOURCE).equals(LoginSource.URL.name())) {

                URL url = null;
                try {
                    url = new URL(prefs.get(PREFS_SSH_PRIVATE_KEY_URL, null));
                } catch (MalformedURLException e) {
                    LOG.warn("Stored URL is not a valid URL. It will be deleted from the settings", e);
                    prefs.remove(PREFS_SSH_PRIVATE_KEY_URL);

                }

                if (url != null) setSshPrivateKeyUrl(url);
            }
        }
    }

    public boolean correctPlanetlabSfaUrn() {
        if (planetlabSfaHrn == null) return false;

        //expected format: <planetlab prefix>.<subauthority>.<username>
        // example:  SFA hrn:   ple.ibbtple.wim_vandemeerssche

        return planetlabSfaHrn.split("\\.").length == 3;
    }

    public boolean correctAuthority() {
        if (authority == null) return false;
        if (authority.getUrl(ServerType.GeniServerRole.PlanetLabSliceRegistry, 1) == null) return false;
        return true;
    }

    public boolean isReadyToFetchCertificate() {
        if (!correctPlanetlabSfaUrn()) return false;
        if (!correctAuthority()) return false;
        if (keypair == null) return false;
        if (privateKey == null) return false;
        if (privateKeyContent == null) return false;
        return true;
    }

    /**
     * Do everything necessary to create a Sfa login from the planetlab info.
     * <p/>
     * This writes details that can be provided as feedback to the user.
     * <p/>
     *
     * @return if fetching of certificate is successful. If successful this UserLoginModel is also a valid login
     */
    public boolean fetchCertificate() throws CertificateFetchException {

        if (!isReadyToFetchCertificate()) {
            throw new CertificateFetchException("Some data needed to fetch the certificate is missing.");
        }

        GeniUrn planetlabUserUrn = sfaHrnToUrn(planetlabSfaHrn);
        userUrn = planetlabUserUrn;
        LOG.trace("Your planetlab Sfa hrn ('{}') is converted to the user URN: '{}'",
                planetlabSfaHrn, planetlabUserUrn.toString());

        certificateChain = null;


        /*
        To login to planetlab using Sfa, we need to create a self signed credential,
        and then use that to request the real credential from planetlab.
        */

        X509Certificate selfSignedCert =
                PlanetlabCertificateFetcher.createSelfSignedCertificate(authority, planetlabSfaHrn, privateKeyContent, keypair);
        if (selfSignedCert == null) {
            throw new CertificateFetchException("Failed to create self signed certificate for initial planetlab connection.");
        }

        // self signed certificate for initial planetlab connection has now been generated.

        PlanetlabSfaRegistryInterface pleReg = new PlanetlabSfaRegistryInterface(logger, jFedPreferences);
        PlanetlabSfaRegistryInterface.SimpleApiCallReply<AnyCredential> cred;
        try {
            //create user
            List<X509Certificate> selfSignedCertAsChain = new ArrayList<X509Certificate>();
            selfSignedCertAsChain.add(selfSignedCert);
            final GeniUser user =
                    new SimpleGeniUser(authority, planetlabUserUrn, selfSignedCertAsChain, privateKey, null, getPrivateKeyFile());
            cred = pleReg.getSelfCredential((SfaConnection) connectionProvider.getConnectionByAuthority(
                            user, authority, PlanetlabSfaRegistryInterface.class),
                    KeyUtil.x509certificateToPem(selfSignedCert),
                    planetlabUserUrn.toString(),
                    ""
            );
        } catch (JFedException e) {
            userUrn = null;
            throw new CertificateFetchException("Failed to call \"GetSelfCredential\" on planetlab server", e);
        }

        if (cred == null || cred.getValue() == null) {
            userUrn = null;
            throw new CertificateFetchException("Call \"GetSelfCredential\" on planetlab server returned empty credential");
        }

        if (!(cred.getValue() instanceof SfaCredential)) {
            userUrn = null;
            throw new CertificateFetchException("Call \"GetSelfCredential\" on planetlab server returned non SFA credential (type=\"" +
                    cred.getValue().getType() + "\" version=\"" + cred.getValue().getVersion() + "\")");
        }

        String gid = ((SfaCredential) cred.getValue()).getTargetGid();
        //gid can contain multiple certs. We keep them all. Not sure if order is correct here...
        //TODO find a way to sort certificates in chain if needed
        certificateChain = KeyUtil.pemToX509CertificateChain(gid);

        //Success! Call "GetSelfCredential" on planetlab server returned our real certificate.

        //all should be ok to login now!
        return true;
    }

    public static class CertificateFetchException extends Exception {
        public CertificateFetchException() {
        }

        public CertificateFetchException(String message) {
            super(message);
        }

        public CertificateFetchException(String message, Throwable cause) {
            super(message, cause);
        }

        public CertificateFetchException(Throwable cause) {
            super(cause);
        }

        public CertificateFetchException(String message, Throwable cause, boolean enableSuppression, boolean writableStackTrace) {
            super(message, cause, enableSuppression, writableStackTrace);
        }
    }

    /**
     * password is not stored internally
     *
     * @return success
     */
    public boolean unlock(String password) {
        return unlock(password.toCharArray());
    }

    /**
     * password is not stored internally
     *
     * @return success
     */
    public boolean unlock(char[] password) {
        if (privateKeyContent == null) return false;

        try {
            this.keypair = KeyUtil.pemToRsaKeyPair(privateKeyContent, password);
            this.privateKey = (RSAPrivateKey) keypair.getPrivate();

            if (keypair == null || privateKey == null)
                LOG.trace("Failed to unlock private key using password");
        } catch (KeyUtil.PEMDecodingException pemDecodingError) {
            LOG.trace("Failed to unlock private key using password", pemDecodingError);
            this.privateKey = null;
        }

        return this.privateKey != null;
    }


    public boolean isPasswordRequired() {
        return passwordRequired;
    }

    public File getSshPrivateKeyFile() {
        return sshPrivateKeyFile;
    }

    @Override
    public File getPrivateKeyFile() {
        return sshPrivateKeyFile;
    }

    public URL getPrivateKeyUrl() {
        return sshPrivateKeyUrl;
    }

    public LoginSource getLoginSource() {
        return loginSource;
    }

    @Override
    public File getCertificateFile() {
        return null;
    }


    public String getPrivateKeyContent() {
        return privateKeyContent;
    }

    public String getSfaHrn() {
        return planetlabSfaHrn;
    }


    @Override
    public String getUserUrnString() {
        return userUrn == null ? null : userUrn.toString();
    }

    @Override
    public GeniUrn getUserUrn() {
        return userUrn;
    }


    @Override
    public PrivateKey getPrivateKey() {
        return privateKey;
    }

    @Override
    public PublicKey getPublicKey() {
        List<X509Certificate> certificateChain = getClientCertificateChain();
        if (certificateChain == null || certificateChain.isEmpty())
            return null;
        return certificateChain.get(0).getPublicKey();
    }

    @Override
    public List<X509Certificate> getClientCertificateChain() {
        return certificateChain;
    }

    @Override
    public SfaAuthority getUserAuthority() {
        return authority;
    }

    public String getUserAuthorityUrn() {
        return authority == null ? null : authority.getUrnString();
    }


    public boolean isValid() {
        if (authority == null) return false;
        if (userUrn == null) return false;
        if (planetlabSfaHrn == null) return false;

        if (loginSource == LoginSource.FILE) {
            if (sshPrivateKeyFile == null) return false;
        } else if (loginSource == LoginSource.URL) {
            if (sshPrivateKeyUrl == null) return false;
        }

        if (privateKeyContent == null) return false;
        if (privateKey == null) return false;
        if (certificateChain == null) return false;
        if (certificateChain.isEmpty()) return false;
        return true;
    }


    @Override
    public GeniUser getLoggedInGeniUser() {
        if (!isValid()) return null;
        return this;
    }

    @Override
    public boolean isUserLoggedIn() {
        return isValid();
    }


    @Override
    public boolean equals(Object o) {
        throw new UnsupportedOperationException();
    }

    @Override
    public int hashCode() {
        throw new UnsupportedOperationException();
    }


    private enum LoginSource {FILE, URL, STRING}
}
