package be.iminds.ilabt.jfed;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.TestCertHelper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.*;
import be.iminds.ilabt.jfed.util.JFedPasswordManager;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import com.google.inject.AbstractModule;
import com.google.inject.Provides;

import javax.inject.Singleton;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.*;

/**
 * Fake connections for testing, with mock replies to requests.
 * <p/>
 * This uses almost every real thing, except for the underlying XmlRpc transport etc.
 */
public class MockHttpRepliesTestCoreModule extends AbstractModule {
    protected static final TestCertHelper DEFAULT_TEST_CERT_HELPER = new TestCertHelper();
    protected final TestCertHelper testCertHelper;
    protected final List<MockSfaConnection.MockXmlRpcReply> mockReplies;


    protected final List<MockSfaConnection.MockXmlRpcHistoryItem> mockXmlRpcHistory =
            Collections.synchronizedList(new ArrayList<MockSfaConnection.MockXmlRpcHistoryItem>());
    protected final List<MockSfaConnection.MockXmlRpcRequestInfo> noMockXmlRpcRequestInfo =
            Collections.synchronizedList(new ArrayList<MockSfaConnection.MockXmlRpcRequestInfo>());

    protected final MockSfaConnection.MockXmlRpcHistoryListener mockXmlRpcHistoryListener =
            new MockSfaConnection.MockXmlRpcHistoryListener() {
                @Override
                public void onMockXmlRpcHistory(MockSfaConnection.MockXmlRpcHistoryItem item) {
                    mockXmlRpcHistory.add(item);
                }
            };

    protected final MockSfaConnection.OnMockXmlRpcListener onNoMockXmlRpcListener =
            new MockSfaConnection.OnMockXmlRpcListener() {
                @Override
                public void onNoMockFor(MockSfaConnection.MockXmlRpcRequestInfo mockXmlRpcRequestInfo) {
                    noMockXmlRpcRequestInfo.add(mockXmlRpcRequestInfo);
                }
            };


    public MockHttpRepliesTestCoreModule(List<MockSfaConnection.MockXmlRpcReply> mockReplies) {
        this.testCertHelper = DEFAULT_TEST_CERT_HELPER;
        this.mockReplies = mockReplies;
    }

    public MockHttpRepliesTestCoreModule() {
        this(new TestCertHelper());
    }

    public MockHttpRepliesTestCoreModule(TestCertHelper testCertHelper) {
        this.testCertHelper = testCertHelper;
        this.mockReplies = new ArrayList<MockSfaConnection.MockXmlRpcReply>();
    }

    public void addMockReply(MockSfaConnection.MockXmlRpcReply mr) {
        mockReplies.add(mr);
    }


    @Override
    protected void configure() {
        bind(JFedTrustStore.class).in(Singleton.class);
        bind(TestCertHelper.class).toInstance(testCertHelper);
        bind(JFedConnectionProvider.class).to(TestSfaConnectionPool.class).in(Singleton.class);

        bind(AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory.class).
                to(AutomaticAggregateManagerWrapper.BasicAutomaticAggregateManagerWrapperFactory.class).in(Singleton.class);

        bind(AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory.class).
                to(AutomaticUserAndSliceApiWrapper.BasicAutomaticUserAndSliceApiWrapperFactory.class).in(Singleton.class);
    }

    private Map<String, MockSfaConnection> conCache = Collections.synchronizedMap(new HashMap<String, MockSfaConnection>());

    public MockSfaConnection createCon(AuthorityListModel authorityListModel, String urlStr) throws JFedException {
        if (conCache.containsKey(urlStr)) return conCache.get(urlStr);

        URL url;
        try {
            url = new URL(urlStr);
        } catch (MalformedURLException e) {
            throw new RuntimeException(e);
        }
        SfaAuthority auth = null;
        for (SfaAuthority a : authorityListModel.getAuthorities()) {
            if (a.getUrls().containsValue(url))
                auth = a;
        }
        assert auth != null;

        MockSfaConnection con;
        try {
            con = new MockSfaConnection(new URL(urlStr), auth, null);
        } catch (MalformedURLException e) {
            throw new RuntimeException("Malformed URL used in test: " + urlStr, e);
        }
        con.setMockReplies(mockReplies);

        con.setNoMockListeners(Collections.singletonList(onNoMockXmlRpcListener));
        con.setHistoryListeners(Collections.singletonList(mockXmlRpcHistoryListener));

        conCache.put(urlStr, con);

        return con;
    }

    @Provides
    public ConnectionBuilderFactory provideConnectionBuilderFactory(final AuthorityListModel authorityListModel) throws JFedException {
        assert authorityListModel != null;
        SfaAuthority userAuth = authorityListModel.getFromAnyUrnExact(TestCertHelper.saUrn);
        SfaAuthority amAuth = authorityListModel.getFromAnyUrnExact(TestCertHelper.am1Urn);
        SfaAuthority am2Auth = authorityListModel.getFromAnyUrnExact(TestCertHelper.am2Urn);
        SfaAuthority scsAuth = authorityListModel.getFromAnyUrnExact(TestCertHelper.scsUrn);
        assert userAuth != null;
        assert amAuth != null;
        assert am2Auth != null;
        assert scsAuth != null;

        return new ConnectionBuilderFactory() {
            @Override
            public ConnectionBuilder createConnectionBuilder() {
                return new ConnectionBuilder() {
                    private String serverUrlStr;

                    @Override
                    public void setUrl(String serverUrlString) {
                        this.serverUrlStr = serverUrlString;
                    }

                    @Override
                    public synchronized void setUrl(URL serverUrl) {
                        //ConnectionBuilder should only be used once. Otherwise, a race condition can cause server connections to the wrong mock server.
                        assert serverUrlStr == null;
                        this.serverUrlStr = serverUrl.toExternalForm();
                    }

                    @Override
                    public synchronized SfaConnection buildSfaConnection() throws JFedException {
                        return createCon(authorityListModel, serverUrlStr);
                    }

                    @Override
                    public HttpConnection buildHttpConnection() throws JFedException {
                        return null;
                    }

                    @Override
                    public ConnectionConfig getCurrentConnectionConfig(ConnectionConfig.Type type) {
                        return null;
                    }

                    @Override
                    public void set(ConnectionConfig connectionConfig) {

                    }

                    @Override
                    public void setHackAllowAllServerCertificates(boolean hackAllowAllServerCertificates) {

                    }

                    @Override
                    public void setDebugInfo(JFedConnection.DebugInfo debugInfo) {
                    }

                    @Override
                    public void setProxy(JFedConnection.ProxyInfo proxyInfo) {
                    }

                    @Override
                    public void useNoAuthentication() {
                    }

                    @Override
                    public void useHttpBasicAuthentication(JFedPasswordManager.LoginInfo login) {
                    }

                    @Override
                    public void useSslClientAuthentication(List<X509Certificate> clientCertChain, PrivateKey privateKey) {
                    }

                    @Override
                    public void useHttp() {
                    }

                    @Override
                    public void useHttps(JFedTrustStore conTrustStore, HandleUntrustedCallback handleUntrustedCallback) {
                    }
                };
            }
        };
    }

    public TestCertHelper getTestCertHelper() {
        return testCertHelper;
    }

    public List<MockSfaConnection.MockXmlRpcHistoryItem> getMockXmlRpcHistory() {
        return mockXmlRpcHistory;
    }

    public List<MockSfaConnection.MockXmlRpcRequestInfo> getNoMockXmlRpcRequestInfo() {
        return noMockXmlRpcRequestInfo;
    }
}
